#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

from uuid import uuid4
from sys import version_info

from proton import *

from . import common


class Test(common.Test):

    def setUp(self):
        self.msg = Message()

    def tearDown(self):
        self.msg = None


class AccessorsTest(Test):

    def _test(self, name, default, values):
        d = getattr(self.msg, name)
        assert d == default, (d, default)
        for v in values:
            setattr(self.msg, name, v)
            gotten = getattr(self.msg, name)
            assert gotten == v, gotten

    def _test_symbol(self, name):
        self._test(name, symbol(None), (symbol(u"abc.123.#$%"), symbol(u"hello.world")))

    def _test_str(self, name):
        self._test(name, None, (u"asdf", u"fdsa", u""))

    def _test_time(self, name):
        self._test(name, 0, (0, 123456789, 987654321))

    def testId(self):
        self._test("id", None, ("bytes", None, 123, u"string", uuid4()))

    def testCorrelationId(self):
        self._test("correlation_id", None, ("bytes", None, 123, u"string", uuid4()))

    def testDurable(self):
        self._test("durable", False, (True, False))

    def testPriority(self):
        self._test("priority", Message.DEFAULT_PRIORITY, range(0, 255))

    def testTtl(self):
        self._test("ttl", 0, range(12345, 54321))

    def testFirstAcquirer(self):
        self._test("first_acquirer", False, (True, False))

    def testDeliveryCount(self):
        self._test("delivery_count", 0, range(0, 1024))

    def testUserId(self):
        self._test("user_id", b"", (b"asdf", b"fdsa", b"asd\x00fdsa", b""))

    def testAddress(self):
        self._test_str("address")

    def testSubject(self):
        self._test_str("subject")

    def testReplyTo(self):
        self._test_str("reply_to")

    def testContentType(self):
        self._test_symbol("content_type")

    def testContentEncoding(self):
        self._test_symbol("content_encoding")

    def testExpiryTime(self):
        self._test_time("expiry_time")

    def testCreationTime(self):
        self._test_time("creation_time")

    def testGroupId(self):
        self._test_str("group_id")

    def testGroupSequence(self):
        self._test("group_sequence", 0, (0, 1, 10, 20, 4294967294, 4294967295))

    def testReplyToGroupId(self):
        self._test_str("reply_to_group_id")


long = int
unicode = str


class CodecTest(Test):

    def testProperties(self):
        self.msg.properties = {'key': 'value'}
        data = self.msg.encode()

        msg2 = Message()
        msg2.decode(data)

        assert msg2.properties['key'] == 'value', msg2.properties['key']

    class MyStringSubclass(unicode):
        def __repr__(self):
            return 'MyStringSubclass(%s)' % unicode.__repr__(self)

    def testStringSubclassPropertyKey(self):
        self.msg.properties = {'abc': 123, CodecTest.MyStringSubclass('def'): 456}
        data = self.msg.encode()

        msg2 = Message()
        msg2.decode(data)

        assert msg2.properties == self.msg.properties
        for k in msg2.properties:
            assert type(k) is unicode, 'non-string key %s %s' % (k, type(k))

    def _testNonStringPropertyKey(self, k):
        self.msg.properties = {k: 'abc'}
        try:
            self.msg.encode()
        except MessageException:
            return
        self.fail('Non-string property key: %s, key type: %s' % (self.msg, type(k)))

    def testNonSequencePropertyKeys(self):
        # AMQP non-string types
        self._testNonStringPropertyKey(None)
        self._testNonStringPropertyKey(True)
        self._testNonStringPropertyKey(byte(1))
        self._testNonStringPropertyKey(ubyte(2))
        self._testNonStringPropertyKey(short(3))
        self._testNonStringPropertyKey(ushort(4))
        self._testNonStringPropertyKey(int32(5))
        self._testNonStringPropertyKey(uint(6))
        self._testNonStringPropertyKey(long(7))
        self._testNonStringPropertyKey(ulong(8))
        self._testNonStringPropertyKey(float32(9.0))
        self._testNonStringPropertyKey(10.0)
        self._testNonStringPropertyKey(decimal32(11))
        self._testNonStringPropertyKey(decimal64(12))
        self._testNonStringPropertyKey(timestamp(1234567890))
        self._testNonStringPropertyKey(uuid4())

    def testCharPropertyKey(self):
        self._testNonStringPropertyKey(char('A'))

    def testDecimal128PropertyKey(self):
        self._testNonStringPropertyKey(decimal128(b'12345'))

    def testSymbolPropertyKey(self):
        self._testNonStringPropertyKey(symbol('abcdef'))

    def testBinaryPropertyKey(self):
        if version_info[0] > 2:
            # Py3: Binary conversions not supported
            self._testNonStringPropertyKey(b'abc123')
        else:
            # Py2: Check for correct key conversion
            self.msg.properties = {'abc': 123, b'def': 456}
            data = self.msg.encode()

            msg2 = Message()
            msg2.decode(data)

            assert msg2.properties == self.msg.properties
            for k in msg2.properties:
                assert type(k) is unicode, 'non-string key %s %s' % (k, type(k))

    def testAnnotationsSymbolicAndUlongKey(self, a=None):
        if a is None:
            a = {symbol('one'): 1, 'two': 2, ulong(3): 'three'}
        self.msg.annotations = a
        data = self.msg.encode()

        msg2 = Message()
        msg2.decode(data)
        # both keys must be symbols
        assert msg2.annotations == a

    def testRoundTrip(self):
        self.msg.id = "asdf"
        self.msg.correlation_id = uuid4()
        self.msg.ttl = 3
        self.msg.priority = 100
        self.msg.address = "address"
        self.msg.subject = "subject"
        self.msg.body = 'Hello World!'

        data = self.msg.encode()

        msg2 = Message()
        msg2.decode(data)

        assert self.msg.id == msg2.id, (self.msg.id, msg2.id)
        assert self.msg.correlation_id == msg2.correlation_id, (self.msg.correlation_id, msg2.correlation_id)
        assert self.msg.ttl == msg2.ttl, (self.msg.ttl, msg2.ttl)
        assert self.msg.priority == msg2.priority, (self.msg.priority, msg2.priority)
        assert self.msg.address == msg2.address, (self.msg.address, msg2.address)
        assert self.msg.subject == msg2.subject, (self.msg.subject, msg2.subject)
        assert self.msg.body == msg2.body, (self.msg.body, msg2.body)

    def testExpiryEncodeAsNull(self):
        self.msg.group_id = "A"  # Force creation and expiry fields to be present
        data = self.msg.encode()

        decoder = Data()

        # Skip past the headers
        consumed = decoder.decode(data)
        decoder.clear()
        data = data[consumed:]

        decoder.decode(data)
        dproperties = decoder.get_py_described()
        # Check we've got the correct described list
        assert dproperties.descriptor == 0x73, (dproperties.descriptor)

        properties = dproperties.value
        assert properties[8] is None, properties[8]

    def testCreationEncodeAsNull(self):
        self.msg.group_id = "A"  # Force creation and expiry fields to be present
        data = self.msg.encode()

        decoder = Data()

        # Skip past the headers
        consumed = decoder.decode(data)
        decoder.clear()
        data = data[consumed:]

        decoder.decode(data)
        dproperties = decoder.get_py_described()
        # Check we've got the correct described list
        assert dproperties.descriptor == 0x73, (dproperties.descriptor)

        properties = dproperties.value
        assert properties[9] is None, properties[9]

    def testGroupSequenceEncodeAsNull(self):
        self.msg.reply_to_group_id = "R"  # Force group_id and group_sequence fields to be present
        data = self.msg.encode()

        decoder = Data()

        # Skip past the headers
        consumed = decoder.decode(data)
        decoder.clear()
        data = data[consumed:]

        decoder.decode(data)
        dproperties = decoder.get_py_described()
        # Check we've got the correct described list
        assert dproperties.descriptor == 0x73, (dproperties.descriptor)

        properties = dproperties.value
        assert properties[10] is None, properties[10]
        assert properties[11] is None, properties[11]

    def testGroupSequenceEncodeAsNonNull(self):
        self.msg.group_id = "G"
        self.msg.reply_to_group_id = "R"  # Force group_id and group_sequence fields to be present
        data = self.msg.encode()

        decoder = Data()

        # Skip past the headers
        consumed = decoder.decode(data)
        decoder.clear()
        data = data[consumed:]

        decoder.decode(data)
        dproperties = decoder.get_py_described()
        # Check we've got the correct described list
        assert dproperties.descriptor == 0x73, (dproperties.descriptor)

        properties = dproperties.value
        assert properties[10] == 'G', properties[10]
        assert properties[11] == 0, properties[11]

    def testDefaultCreationExpiryDecode(self):
        # This is a message with everything filled explicitly as null or zero in LIST32 HEADER and PROPERTIES lists
        data = b'\x00\x53\x70\xd0\x00\x00\x00\x0a\x00\x00\x00\x05\x42\x40\x40\x42\x52\x00\x00\x53\x73\xd0\x00\x00\x00\x12\x00\x00\x00\x0d\x40\x40\x40\x40\x40\x40\x40\x40\x40\x40\x40\x52\x00\x40'
        msg2 = Message()
        msg2.decode(data)
        assert msg2.expiry_time == 0, (msg2.expiry_time)
        assert msg2.creation_time == 0, (msg2.creation_time)

        # The same message with LIST8s instead
        data = b'\x00\x53\x70\xc0\x07\x05\x42\x40\x40\x42\x52\x00\x00\x53\x73\xc0\x0f\x0d\x40\x40\x40\x40\x40\x40\x40\x40\x40\x40\x40\x52\x00\x40'
        msg3 = Message()
        msg3.decode(data)
        assert msg2.expiry_time == 0, (msg2.expiry_time)
        assert msg2.creation_time == 0, (msg2.creation_time)

        # Minified message with zero length HEADER and PROPERTIES lists
        data = b'\x00\x53\x70\x45' b'\x00\x53\x73\x45'
        msg4 = Message()
        msg4.decode(data)
        assert msg2.expiry_time == 0, (msg2.expiry_time)
        assert msg2.creation_time == 0, (msg2.creation_time)

    def testDefaultPriorityEncode(self):
        assert self.msg.priority == 4, (self.msg.priority)
        self.msg.ttl = 0.003  # field after priority, so forces priority to be present
        data = self.msg.encode()

        decoder = Data()
        decoder.decode(data)

        dheaders = decoder.get_py_described()
        # Check we've got the correct described list
        assert dheaders.descriptor == 0x70, (dheaders.descriptor)

        # Check that the priority field (second field) is encoded as null
        headers = dheaders.value
        assert headers[1] is None, (headers[1])

    def testDefaultPriorityDecode(self):
        # This is a message with everything filled explicitly as null or zero in LIST32 HEADER and PROPERTIES lists
        data = b'\x00\x53\x70\xd0\x00\x00\x00\x0a\x00\x00\x00\x05\x42\x40\x40\x42\x52\x00\x00\x53\x73\xd0\x00\x00\x00\x22\x00\x00\x00\x0d\x40\x40\x40\x40\x40\x40\x40\x40\x83\x00\x00\x00\x00\x00\x00\x00\x00\x83\x00\x00\x00\x00\x00\x00\x00\x00\x40\x52\x00\x40'
        msg2 = Message()
        msg2.decode(data)
        assert msg2.priority == 4, (msg2.priority)

        # The same message with LIST8s instead
        data = b'\x00\x53\x70\xc0\x07\x05\x42\x40\x40\x42\x52\x00\x00\x53\x73\xc0\x1f\x0d\x40\x40\x40\x40\x40\x40\x40\x40\x83\x00\x00\x00\x00\x00\x00\x00\x00\x83\x00\x00\x00\x00\x00\x00\x00\x00\x40\x52\x00\x40'
        msg3 = Message()
        msg3.decode(data)
        assert msg3.priority == 4, (msg3.priority)

        # Minified message with zero length HEADER and PROPERTIES lists
        data = b'\x00\x53\x70\x45' b'\x00\x53\x73\x45'
        msg4 = Message()
        msg4.decode(data)
        assert msg4.priority == 4, (msg4.priority)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License
#

"""
PROTON-1709 [python] ApplicationEvent causing memory growth
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gc
import platform
import threading
import unittest

import proton
from proton.handlers import MessagingHandler
from proton.reactor import Container, ApplicationEvent, EventInjector


class Program(MessagingHandler):
    def __init__(self, injector):
        self.injector = injector
        self.counter = 0
        self.on_start_ = threading.Event()

    def on_reactor_init(self, event):
        event.reactor.selectable(self.injector)
        self.on_start_.set()

    def on_count_up(self, event):
        self.counter += 1
        gc.collect()

    def on_done(self, event):
        event.subject.stop()


class Proton1709Test(unittest.TestCase):
    @unittest.skipIf(platform.system() == 'Windows', "TODO jdanek: Test is broken on Windows")
    def test_application_event_no_object_leaks(self):
        event_types_count = len(proton.EventType.TYPES)

        injector = EventInjector()
        p = Program(injector)
        c = Container(p)
        t = threading.Thread(target=c.run)
        t.start()

        p.on_start_.wait()

        object_counts = []

        gc.collect()
        object_counts.append(len(gc.get_objects()))

        for i in range(100):
            injector.trigger(ApplicationEvent("count_up"))

        gc.collect()
        object_counts.append(len(gc.get_objects()))

        self.assertEqual(len(proton.EventType.TYPES), event_types_count + 1)

        injector.trigger(ApplicationEvent("done", subject=c))
        self.assertEqual(len(proton.EventType.TYPES), event_types_count + 2)

        t.join()

        gc.collect()
        object_counts.append(len(gc.get_objects()))

        self.assertEqual(p.counter, 100)

        self.assertTrue(object_counts[1] - object_counts[0] <= 220,
                        "Object counts should not be increasing too fast: {0}".format(object_counts))
        self.assertTrue(object_counts[2] - object_counts[0] <= 10,
                        "No objects should be leaking at the end: {0}".format(object_counts))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

import errno
import socket
import select
import time

from typing import TYPE_CHECKING, Tuple, List

if TYPE_CHECKING:
    from proton._selectable import Selectable

PN_INVALID_SOCKET = -1


class IO(object):

    @staticmethod
    def _setupsocket(s: socket.socket) -> None:
        s.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, True)
        s.setblocking(False)

    @staticmethod
    def close(s: socket.socket) -> None:
        s.close()

    @staticmethod
    def listen(host, port) -> socket.socket:
        s = socket.socket()
        IO._setupsocket(s)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, True)
        s.bind((host, port))
        s.listen(10)
        return s

    @staticmethod
    def accept(s: socket.socket):
        n = s.accept()
        IO._setupsocket(n[0])
        return n

    @staticmethod
    def connect(addr) -> socket.socket:
        s = socket.socket(addr[0], addr[1], addr[2])
        IO._setupsocket(s)
        try:
            s.connect(addr[4])
        except socket.error as e:
            if e.errno not in (errno.EINPROGRESS, errno.EWOULDBLOCK, errno.EAGAIN):
                raise
        return s

    @staticmethod
    def select(*args, **kwargs):
        return select.select(*args, **kwargs)

    @staticmethod
    def sleep(t: float) -> None:
        time.sleep(t)
        return

    class Selector(object):

        def __init__(self) -> None:
            self._selectables = set()
            self._reading = set()
            self._writing = set()
            self._deadline = None

        def add(self, selectable: 'Selectable') -> None:
            self._selectables.add(selectable)
            if selectable.reading:
                self._reading.add(selectable)
            if selectable.writing:
                self._writing.add(selectable)
            if selectable.deadline:
                if self._deadline is None:
                    self._deadline = selectable.deadline
                else:
                    self._deadline = min(selectable.deadline, self._deadline)

        def remove(self, selectable: 'Selectable') -> None:
            self._selectables.discard(selectable)
            self._reading.discard(selectable)
            self._writing.discard(selectable)
            self.update_deadline()

        @property
        def selectables(self) -> int:
            return len(self._selectables)

        def update_deadline(self) -> None:
            for sel in self._selectables:
                if sel.deadline:
                    if self._deadline is None:
                        self._deadline = sel.deadline
                    else:
                        self._deadline = min(sel.deadline, self._deadline)

        def update(self, selectable: 'Selectable') -> None:
            self._reading.discard(selectable)
            self._writing.discard(selectable)
            if selectable.reading:
                self._reading.add(selectable)
            if selectable.writing:
                self._writing.add(selectable)
            self.update_deadline()

        def select(self, timeout: float) -> Tuple[List, List, List]:

            def select_inner(timeout):
                # This inner select adds the writing fds to the exception fd set
                # because Windows returns connected fds in the exception set not the
                # writable set
                r = self._reading
                w = self._writing

                now = time.time()

                # No timeout or deadline
                if timeout is None and self._deadline is None:
                    return IO.select(r, w, w)

                if timeout is None:
                    t = max(0, self._deadline - now)
                    return IO.select(r, w, w, t)

                if self._deadline is None:
                    return IO.select(r, w, w, timeout)

                t = max(0, min(timeout, self._deadline - now))
                if len(r) == 0 and len(w) == 0:
                    if t > 0:
                        IO.sleep(t)
                    return ([], [], [])

                return IO.select(r, w, w, t)

            # Need to allow for signals interrupting us on Python 2
            # In this case the signal handler could have messed up our internal state
            # so don't retry just return with no handles.
            try:
                r, w, ex = select_inner(timeout)
            except select.error as e:
                if e.errno != errno.EINTR:
                    raise
                r, w, ex = ([], [], [])

            # For windows non blocking connect we get exception not writable so add exceptions to writable
            w += ex

            # Calculate timed out selectables
            now = time.time()
            t = [s for s in self._selectables if s.deadline and now > s.deadline]
            self._deadline = None
            self.update_deadline()
            return r, w, t



#!/usr/bin/env python3
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import sqlite3
import sys

if len(sys.argv) < 3:
    print("Usage: %s [init|insert|list] db" % sys.argv[0])
else:
    conn = sqlite3.connect(sys.argv[2])
    with conn:
        if sys.argv[1] == "init":
            conn.execute("DROP TABLE IF EXISTS records")
            conn.execute("CREATE TABLE records(id INTEGER PRIMARY KEY AUTOINCREMENT, description TEXT)")
            conn.commit()
        elif sys.argv[1] == "list":
            cursor = conn.cursor()
            cursor.execute("SELECT * FROM records")
            rows = cursor.fetchall()
            for r in rows:
                print(r)
        elif sys.argv[1] == "insert":
            while True:
                l = sys.stdin.readline()
                if not l:
                    break
                conn.execute("INSERT INTO records(description) VALUES (?)", (l.rstrip(),))
            conn.commit()
        else:
            print("Unrecognised command: %s" % sys.argv[1])



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""
Parsers for XML to dom.
"""
from __future__ import absolute_import

import xml.sax.handler
from .dom import *


class Parser:

    def __init__(self):
        self.tree = Tree()
        self.node = self.tree
        self.nodes = []

    def line(self, id, lineno, colno):
        while self.nodes:
            n = self.nodes.pop()
            n._line(id, lineno, colno)

    def add(self, node):
        self.node.add(node)
        self.nodes.append(node)

    def start(self, name, attrs):
        tag = Tag(name, *attrs)
        self.add(tag)
        self.node = tag

    def end(self, name):
        self.balance(name)
        self.node = self.node.parent

    def data(self, data):
        children = self.node.children
        if children and isinstance(children[-1], Data):
            children[-1].data += data
        else:
            self.add(Data(data))

    def comment(self, comment):
        self.add(Comment(comment))

    def entity(self, ref):
        self.add(Entity(ref))

    def character(self, ref):
        self.add(Character(ref))

    def balance(self, name=None):
        while self.node != self.tree and name != self.node.name:
            self.node.parent.extend(self.node.children)
            del self.node.children[:]
            self.node.singleton = True
            self.node = self.node.parent


class XMLParser(xml.sax.handler.ContentHandler):

    def __init__(self):
        self.parser = Parser()
        self.locator = None

    def line(self):
        if self.locator is not None:
            self.parser.line(self.locator.getSystemId(),
                             self.locator.getLineNumber(),
                             self.locator.getColumnNumber())

    def setDocumentLocator(self, locator):
        self.locator = locator

    def startElement(self, name, attrs):
        self.parser.start(name, attrs.items())
        self.line()

    def endElement(self, name):
        self.parser.end(name)
        self.line()

    def characters(self, content):
        self.parser.data(content)
        self.line()

    def skippedEntity(self, name):
        self.parser.entity(name)
        self.line()



from __future__ import absolute_import
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import time
import sys
import json
from .common import Test, SkipTest, TestServer, free_tcp_port, ensureCanTestExtendedSASL
from proton import SSLDomain
from proton.reactor import Container
from proton.handlers import MessagingHandler
from .ssl import _testpath


def write_connect_conf(obj):
    with open('connect.json', 'w') as outfile:
        json.dump(obj, outfile)


class Server(MessagingHandler):
    def __init__(self, expected_user=None, scheme='amqps'):
        super(Server, self).__init__()
        self.port = free_tcp_port()
        self.scheme = scheme
        self.url = '%s://localhost:%i' % (self.scheme, self.port)
        self.expected_user = expected_user
        self.verified_user = False

    def on_start(self, event):
        self.listener = event.container.listen(self.url)

    def on_connection_opening(self, event):
        if self.expected_user:
            assert event.connection.transport.user == self.expected_user
            self.verified_user = True

    def on_connection_closing(self, event):
        event.connection.close()
        self.listener.close()


class Client(MessagingHandler):
    def __init__(self):
        super(Client, self).__init__()
        self.opened = False

    def on_connection_opened(self, event):
        self.opened = True
        event.connection.close()


class ConnectConfigTest(Test):
    def test_port(self):
        ensureCanTestExtendedSASL()
        server = Server(scheme='amqp')
        container = Container(server)
        client = Client()
        write_connect_conf({'port': server.port, 'scheme': 'amqp'})
        container.connect(handler=client, reconnect=False)
        container.run()
        assert client.opened

    def test_user(self):
        ensureCanTestExtendedSASL()
        user = 'user@proton'
        password = 'password'
        server = Server(user)
        container = Container(server)
        container.ssl.server.set_credentials(_testpath('server-certificate.pem'),
                                             _testpath('server-private-key.pem'),
                                             'server-password')
        client = Client()
        config = {
            'user': user,
            'password': password,
            'port': server.port,
            'tls': {
                'verify': False,
                'ca': _testpath('ca-certificate.pem')
            }
        }
        write_connect_conf(config)
        container.connect(handler=client, reconnect=False)
        container.run()
        assert client.opened
        assert server.verified_user

    def test_ssl(self):
        ensureCanTestExtendedSASL()
        server = Server(scheme='amqps')
        container = Container(server)
        container.ssl.server.set_credentials(_testpath('server-certificate-lh.pem'),
                                             _testpath('server-private-key-lh.pem'),
                                             'server-password')
        client = Client()
        config = {
            'scheme': 'amqps',
            'port': server.port,
            'tls': {
                'verify': True,
                'ca': _testpath('ca-certificate.pem')
            }
        }
        write_connect_conf(config)
        container.connect(handler=client, reconnect=False)
        container.run()
        assert client.opened

    def test_ssl_external(self):
        ensureCanTestExtendedSASL()
        server = Server(scheme='amqps')
        container = Container(server)
        container.ssl.server.set_credentials(_testpath('server-certificate-lh.pem'),
                                             _testpath('server-private-key-lh.pem'),
                                             'server-password')
        container.ssl.server.set_trusted_ca_db(_testpath('ca-certificate.pem'))
        container.ssl.server.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                     _testpath('ca-certificate.pem'))

        client = Client()
        config = {
            'scheme': 'amqps',
            'port': server.port,
            'sasl': {
                'mechanisms': 'EXTERNAL'
            },
            'tls': {
                'cert': _testpath('client-certificate.pem'),
                'key': _testpath('client-private-key-no-password.pem'),
                'ca': _testpath('ca-certificate.pem'),
                'verify': True
            }
        }
        write_connect_conf(config)
        container.connect(handler=client, reconnect=False)
        container.run()
        assert client.opened

    def test_ssl_plain(self):
        ensureCanTestExtendedSASL()
        user = 'user@proton'
        password = 'password'
        server = Server(expected_user=user, scheme='amqps')
        container = Container(server)
        container.ssl.server.set_credentials(_testpath('server-certificate-lh.pem'),
                                             _testpath('server-private-key-lh.pem'),
                                             'server-password')
        container.ssl.server.set_trusted_ca_db(_testpath('ca-certificate.pem'))
        container.ssl.server.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                     _testpath('ca-certificate.pem'))

        client = Client()
        config = {
            'scheme': 'amqps',
            'port': server.port,
            'user': user,
            'password': password,
            'sasl': {
                'mechanisms': 'PLAIN'
            },
            'tls': {
                'cert': _testpath('client-certificate.pem'),
                'key': _testpath('client-private-key-no-password.pem'),
                'ca': _testpath('ca-certificate.pem'),
                'verify': True
            }
        }
        write_connect_conf(config)
        container.connect(handler=client, reconnect=False)
        container.run()
        assert client.opened



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
import mllib
import os
import sys

doc = mllib.xml_parse(os.path.join(os.path.dirname(__file__), "transport.xml"))
mdoc = mllib.xml_parse(os.path.join(os.path.dirname(__file__), "messaging.xml"))
tdoc = mllib.xml_parse(os.path.join(os.path.dirname(__file__), "transactions.xml"))
sdoc = mllib.xml_parse(os.path.join(os.path.dirname(__file__), "security.xml"))


def eq(attr, value):
    return lambda nd: nd[attr] == value


TYPEStmp = doc.query["amqp/section/type", eq("@class", "composite")] + \
    mdoc.query["amqp/section/type", eq("@class", "composite")] + \
    tdoc.query["amqp/section/type", eq("@class", "composite")] + \
    sdoc.query["amqp/section/type", eq("@class", "composite")] + \
    mdoc.query["amqp/section/type", eq("@provides", "section")]
TYPES = []
for ty in TYPEStmp:
    if ty not in TYPES:
        TYPES.append(ty)
RESTRICTIONS = {}
COMPOSITES = {}

for type in doc.query["amqp/section/type"] + mdoc.query["amqp/section/type"] + \
        sdoc.query["amqp/section/type"] + tdoc.query["amqp/section/type"]:

    source = type["@source"]
    if source:
        RESTRICTIONS[type["@name"]] = source
    if type["@class"] == "composite":
        COMPOSITES[type["@name"]] = type


def resolve(name):
    if name in RESTRICTIONS:
        return resolve(RESTRICTIONS[name])
    else:
        return name


TYPEMAP = {
    "boolean": ("bool", "", ""),
    "binary": ("pn_binary_t", "*", ""),
    "string": ("wchar_t", "*", ""),
    "symbol": ("char", "*", ""),
    "ubyte": ("uint8_t", "", ""),
    "ushort": ("uint16_t", "", ""),
    "uint": ("uint32_t", "", ""),
    "ulong": ("uint64_t", "", ""),
    "timestamp": ("uint64_t", "", ""),
    "list": ("pn_list_t", "*", ""),
    "map": ("pn_map_t", "*", ""),
    "box": ("pn_box_t", "*", ""),
    "*": ("pn_object_t", "*", "")
}

CONSTRUCTORS = {
    "boolean": "boolean",
    "string": "string",
    "symbol": "symbol",
    "ubyte": "ubyte",
    "ushort": "ushort",
    "uint": "uint",
    "ulong": "ulong",
    "timestamp": "ulong"
}

NULLABLE = set(["string", "symbol"])


def fname(field):
    return field["@name"].replace("-", "_")


def tname(t):
    return t["@name"].replace("-", "_")


def multi(f):
    return f["@multiple"] == "true"


def ftype(field):
    if multi(field):
        return "list"
    elif field["@type"] in COMPOSITES:
        return "box"
    else:
        return resolve(field["@type"]).replace("-", "_")


def fconstruct(field, expr):
    type = ftype(field)
    if type in CONSTRUCTORS:
        result = "pn_%s(mem, %s)" % (CONSTRUCTORS[type], expr)
        if type in NULLABLE:
            result = "%s ? %s : NULL" % (expr, result)
    else:
        result = expr
    if multi(field):
        result = "pn_box(mem, pn_boolean(mem, true), %s)" % result
    return result


def declaration(field):
    name = fname(field)
    type = ftype(field)
    t, pre, post = TYPEMAP.get(type, (type, "", ""))
    return t, "%s%s%s" % (pre, name, post)


def field_kw(field):
    return fname(field).upper()



import argparse
import json
import os
import re
import sys

from pycparser import c_ast, parse_file

from typing import Any, Dict, List, Set, Tuple, Union


def strip_quotes(arg: str) -> str:
    if arg[0] != '"' or arg[-1:] != '"':
            raise Exception(arg)
    return arg[1:-1]


def find_function_calls(file: str, name: str, includes: List[str], defines: List[str]) -> List[List[Any]]:
    class FillFinder(c_ast.NodeVisitor):
        def __init__(self):
            self._name = name
            self.result = []

        def visit_FuncCall(self, node):
            if node.name.name == self._name:
                r = []
                for e in node.args.exprs:
                    r.append(e)
                self.result.append(r)

    include_args = [f'-I{d}' for d in includes]
    define_args = [f'-D{d}' for d in defines]
    ast = parse_file(file, use_cpp=True,
                     cpp_args=[
                         *include_args,
                         *define_args
                     ])
    ff = FillFinder()
    ff.visit(ast)
    return ff.result


c_defines = ['GENERATE_CODEC_CODE',
             'NDEBUG',
             '__attribute__(X)=',
             '__asm__(X)=',
             '__inline=',
             '__extension__=',
             '__restrict=',
             '__builtin_va_list=int']


def find_fill_specs(c_includes, pn_source):
    amqp_calls = find_function_calls(os.path.join(pn_source, 'c/src/core/transport.c'), 'pn_fill_performative',
                                     c_includes, c_defines)
    sasl_calls = find_function_calls(os.path.join(pn_source, 'c/src/sasl/sasl.c'), 'pn_fill_performative',
                                     c_includes, c_defines)
    message_calls = find_function_calls(os.path.join(pn_source, 'c/src/core/message.c'), 'pn_data_fill',
                                        c_includes, c_defines)
    fill_spec_args = [c[1] for c in amqp_calls]
    fill_spec_args += [c[1] for c in sasl_calls]
    fill_spec_args += [c[1] for c in message_calls]
    fill_specs: Set[str] = \
        {strip_quotes(e.value) for e in fill_spec_args if type(e) is c_ast.Constant and e.type == 'string'}
    return fill_specs


def find_scan_specs(c_includes, pn_source):
    amqp_calls = find_function_calls(os.path.join(pn_source, 'c/src/core/transport.c'), 'pn_data_scan',
                                     c_includes, c_defines)
    message_calls = find_function_calls(os.path.join(pn_source, 'c/src/core/message.c'),'pn_data_scan',
                                        c_includes, c_defines)
    scan_spec_args = [c[1] for c in amqp_calls]
    scan_spec_args += [c[1] for c in message_calls]
    # Only try to generate code for constant strings
    scan_specs: Set[str] = \
        {strip_quotes(e.value) for e in scan_spec_args if type(e) is c_ast.Constant and e.type == 'string'}
    return scan_specs


def main():
    argparser = argparse.ArgumentParser(description='Find scan/fill specs in proton code for code generation')
    argparser.add_argument('-o', '--output', help='output json file', type=str, required=True)
    argparser.add_argument('-s', '--source', help='proton source directory', type=str, required=True)
    argparser.add_argument('-b', '--build', help='proton build directory', type=str, required=True)

    args = argparser.parse_args()

    pn_source = args.source
    pn_build = args.build
    json_filename = args.output

    c_includes = [f'{os.path.join(pn_build, "c/include")}',
                  f'{os.path.join(pn_build, "c/src")}']
    fill_specs = find_fill_specs(c_includes, pn_source)
    scan_specs = find_scan_specs(c_includes, pn_source)
    with open(json_filename, 'w') as file:
        json.dump({'fill_specs': sorted(list(fill_specs)), 'scan_specs': sorted(list(scan_specs))}, file, indent=2)


if __name__ == '__main__':
    sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
    sys.exit(main())



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

import sys
from uuid import uuid4

from proton import *

from . import common


class Test(common.Test):

    def setUp(self):
        self.data = Data()

    def tearDown(self):
        self.data = None


class DataTest(Test):

    def testTopLevelNext(self):
        assert self.data.next() is None
        self.data.put_null()
        self.data.put_bool(False)
        self.data.put_int(0)
        assert self.data.next() is None
        self.data.rewind()
        assert self.data.next() == Data.NULL
        assert self.data.next() == Data.BOOL
        assert self.data.next() == Data.INT
        assert self.data.next() is None

    def testNestedNext(self):
        assert self.data.next() is None
        self.data.put_null()
        assert self.data.next() is None
        self.data.put_list()
        assert self.data.next() is None
        self.data.put_bool(False)
        assert self.data.next() is None
        self.data.rewind()
        assert self.data.next() is Data.NULL
        assert self.data.next() is Data.LIST
        self.data.enter()
        assert self.data.next() is None
        self.data.put_ubyte(0)
        assert self.data.next() is None
        self.data.put_uint(0)
        assert self.data.next() is None
        self.data.put_int(0)
        assert self.data.next() is None
        self.data.exit()
        assert self.data.next() is Data.BOOL
        assert self.data.next() is None

        self.data.rewind()
        assert self.data.next() is Data.NULL
        assert self.data.next() is Data.LIST
        assert self.data.enter()
        assert self.data.next() is Data.UBYTE
        assert self.data.next() is Data.UINT
        assert self.data.next() is Data.INT
        assert self.data.next() is None
        assert self.data.exit()
        assert self.data.next() is Data.BOOL
        assert self.data.next() is None

    def testEnterExit(self):
        assert self.data.next() is None
        assert not self.data.enter()
        self.data.put_list()
        assert self.data.enter()
        assert self.data.next() is None
        self.data.put_list()
        assert self.data.enter()
        self.data.put_list()
        assert self.data.enter()
        assert self.data.exit()
        assert self.data.get_list() == 0
        assert self.data.exit()
        assert self.data.get_list() == 1
        assert self.data.exit()
        assert self.data.get_list() == 1
        assert not self.data.exit()
        assert self.data.get_list() == 1
        assert self.data.next() is None

        self.data.rewind()
        assert self.data.next() is Data.LIST
        assert self.data.get_list() == 1
        assert self.data.enter()
        assert self.data.next() is Data.LIST
        assert self.data.get_list() == 1
        assert self.data.enter()
        assert self.data.next() is Data.LIST
        assert self.data.get_list() == 0
        assert self.data.enter()
        assert self.data.next() is None
        assert self.data.exit()
        assert self.data.get_list() == 0
        assert self.data.exit()
        assert self.data.get_list() == 1
        assert self.data.exit()
        assert self.data.get_list() == 1
        assert not self.data.exit()

    def put(self, putter, v):
        """More informative exception from putters, include bad value"""
        try:
            putter(v)
        except Exception:
            etype, value, trace = sys.exc_info()
            v = etype("%s(%r): %s" % (putter.__name__, v, value))
            if trace is None:
                raise v
            else:
                raise v.with_traceback(trace)

        return putter

    # (bits, signed) for each integer type
    INT_TYPES = {
        "byte": (8, True),
        "ubyte": (8, False),
        "short": (16, True),
        "ushort": (16, False),
        "int": (32, True),
        "uint": (32, False),
        "long": (64, True),
        "ulong": (64, False)
    }

    def int_values(self, dtype):
        """Set of test values for integer type dtype, include extreme and medial values"""
        bits, signed = self.INT_TYPES[dtype]
        values = [0, 1, 2, 5, 42]
        if signed:
            min, max = -2**(bits - 1), 2**(bits - 1) - 1
            values.append(max // 2)
            values += [-i for i in values if i]
            values += [min, max]
        else:
            max = 2**(bits) - 1
            values += [max // 2, max]
        return sorted(values)

    def _testArray(self, dtype, descriptor, atype, *values):
        if dtype:
            dTYPE = getattr(self.data, dtype.upper())
        aTYPE = getattr(self.data, atype.upper())
        self.data.put_array(dtype is not None, aTYPE)
        self.data.enter()
        if dtype is not None:
            putter = getattr(self.data, "put_%s" % dtype)
            self.put(putter, descriptor)
        putter = getattr(self.data, "put_%s" % atype)
        for v in values:
            self.put(putter, v)
        self.data.exit()
        self.data.rewind()
        assert self.data.next() == Data.ARRAY
        count, described, type = self.data.get_array()
        assert count == len(values), count
        if dtype is None:
            assert described == False
        else:
            assert described
        assert type == aTYPE, type
        assert self.data.enter()
        if described:
            assert self.data.next() == dTYPE
            getter = getattr(self.data, "get_%s" % dtype)
            gotten = getter()
            assert gotten == descriptor, gotten
        if values:
            getter = getattr(self.data, "get_%s" % atype)
            for v in values:
                assert self.data.next() == aTYPE
                gotten = getter()
                assert gotten == v, gotten
        assert self.data.next() is None
        assert self.data.exit()

    def testStringArray(self):
        self._testArray(None, None, "string", "one", "two", "three")

    def testDescribedStringArray(self):
        self._testArray("symbol", "url", "string", "one", "two", "three")

    def _test_int_array(self, atype):
        self._testArray(None, None, atype, *self.int_values(atype))

    def testByteArray(self): self._test_int_array("byte")
    def testUbyteArray(self): self._test_int_array("ubyte")
    def testShortArray(self): self._test_int_array("short")
    def testUshortArray(self): self._test_int_array("ushort")
    def testIntArray(self): self._test_int_array("int")
    def testUintArray(self): self._test_int_array("uint")
    def testLongArray(self): self._test_int_array("long")
    def testUlongArray(self): self._test_int_array("ulong")

    def testUUIDArray(self):
        self._testArray(None, None, "uuid", uuid4(), uuid4(), uuid4())

    def testEmptyArray(self):
        self._testArray(None, None, "null")

    def testDescribedEmptyArray(self):
        self._testArray("long", 0, "null")

    def testPropertyDict(self):
        a = PropertyDict(one=1, two=2, three=3)
        b = PropertyDict({'one': 1, 'two': 2, 'three': 3})
        c = PropertyDict(zip(['one', 'two', 'three'], [1, 2, 3]))
        d = PropertyDict([('two', 2), ('one', 1), ('three', 3)])
        e = PropertyDict({symbol('three'): 3, symbol('one'): 1, symbol('two'): 2})
        f = PropertyDict(a)
        g = PropertyDict()
        g['one'] = 1
        g[symbol('two')] = 2
        g['three'] = 3
        assert a == b == c == d == e == f == g
        for k in a.keys():
            assert isinstance(k, symbol)
        self.assertRaises(KeyError, AnnotationDict, {'one': 1, None: 'none'})
        self.assertRaises(KeyError, AnnotationDict, {'one': 1, 1.23: 4})

    def testPropertyDictNoRaiseError(self):
        a = PropertyDict(one=1, two=2, three=3, raise_on_error=False)
        a[4] = 'four'
        b = PropertyDict({'one': 1, 'two': 2, 'three': 3, 4: 'four'}, raise_on_error=False)
        c = PropertyDict(zip(['one', 'two', 'three', 4], [1, 2, 3, 'four']), raise_on_error=False)
        d = PropertyDict([('two', 2), ('one', 1), ('three', 3), (4, 'four')], raise_on_error=False)
        e = PropertyDict({4: 'four', symbol('three'): 3, symbol('one'): 1, symbol('two'): 2}, raise_on_error=False)
        f = PropertyDict(a, raise_on_error=False)
        g = PropertyDict(raise_on_error=False)
        g['one'] = 1
        g[4] = 'four'
        g[symbol('two')] = 2
        g['three'] = 3
        assert a == b == c == d == e == f == g

    def testAnnotationDict(self):
        # AnnotationMap c'tor calls update(), so this method is also covered
        a = AnnotationDict(one=1, two=2, three=3)
        a[ulong(4)] = 'four'
        b = AnnotationDict({'one': 1, 'two': 2, 'three': 3, ulong(4): 'four'})
        c = AnnotationDict(zip(['one', 'two', 'three', ulong(4)], [1, 2, 3, 'four']))
        d = AnnotationDict([('two', 2), ('one', 1), ('three', 3), (ulong(4), 'four')])
        e = AnnotationDict({symbol('three'): 3, ulong(4): 'four', symbol('one'): 1, symbol('two'): 2})
        f = AnnotationDict(a)
        g = AnnotationDict()
        g[ulong(4)] = 'four'
        g['one'] = 1
        g[symbol('two')] = 2
        g['three'] = 3
        assert a == b == c == d == e == f == g
        for k in a.keys():
            assert isinstance(k, (symbol, ulong))
        self.assertRaises(KeyError, AnnotationDict, {'one': 1, None: 'none'})
        self.assertRaises(KeyError, AnnotationDict, {'one': 1, 1.23: 4})

    def testAnnotationDictNoRaiseError(self):
        a = AnnotationDict(one=1, two=2, three=3, raise_on_error=False)
        a[ulong(4)] = 'four'
        a[5] = 'five'
        b = AnnotationDict({'one': 1, 'two': 2, 'three': 3, ulong(4): 'four', 5: 'five'}, raise_on_error=False)
        c = AnnotationDict(zip(['one', 'two', 'three', ulong(4), 5], [1, 2, 3, 'four', 'five']), raise_on_error=False)
        d = AnnotationDict([('two', 2), ('one', 1), ('three', 3),
                            (ulong(4), 'four'), (5, 'five')], raise_on_error=False)
        e = AnnotationDict({5: 'five', symbol('three'): 3, ulong(4): 'four',
                            symbol('one'): 1, symbol('two'): 2}, raise_on_error=False)
        f = AnnotationDict(a, raise_on_error=False)
        g = AnnotationDict(raise_on_error=False)
        g[ulong(4)] = 'four'
        g['one'] = 1
        g[symbol('two')] = 2
        g[5] = 'five'
        g['three'] = 3
        assert a == b == c == d == e == f == g

    def testSymbolList(self):
        a = SymbolList(['one', 'two', 'three'])
        b = SymbolList([symbol('one'), symbol('two'), symbol('three')])
        c = SymbolList()
        c.append('one')
        c.extend([symbol('two'), 'three'])
        d1 = SymbolList(['one'])
        d2 = SymbolList(['two', symbol('three')])
        d = d1 + d2
        e = SymbolList(['one'])
        e += SymbolList(['two', symbol('three')])
        f = SymbolList(['one', 'hello', 'goodbye'])
        f[1] = symbol('two')
        f[2] = 'three'
        g = SymbolList(a)
        assert a == b == c == d == e == f == g
        for v in a:
            assert isinstance(v, symbol)
        self.assertRaises(TypeError, SymbolList, ['one', None])
        self.assertRaises(TypeError, SymbolList, ['one', 2])
        self.assertRaises(TypeError, SymbolList, ['one', ['two']])
        self.assertRaises(TypeError, SymbolList, ['one', {'two': 3}])

    def testSymbolListNoRaiseError(self):
        a = SymbolList(['one', 'two', 'three', 4], raise_on_error=False)
        b = SymbolList([symbol('one'), symbol('two'), symbol('three'), 4], raise_on_error=False)
        c = SymbolList(raise_on_error=False)
        c.append('one')
        c.extend([symbol('two'), 'three', 4])
        d1 = SymbolList(['one'], raise_on_error=False)
        d2 = SymbolList(['two', symbol('three'), 4], raise_on_error=False)
        d = d1 + d2
        e = SymbolList(['one'], raise_on_error=False)
        e += SymbolList(['two', symbol('three'), 4], raise_on_error=False)
        f = SymbolList(['one', 'hello', 'goodbye', 'what?'], raise_on_error=False)
        f[1] = symbol('two')
        f[2] = 'three'
        f[3] = 4
        g = SymbolList(a, raise_on_error=False)
        assert a == b == c == d == e == f == g

    def _test(self, dtype, *values, **kwargs):
        eq = kwargs.get("eq", lambda x, y: x == y)
        ntype = getattr(Data, dtype.upper())
        putter = getattr(self.data, "put_%s" % dtype)
        getter = getattr(self.data, "get_%s" % dtype)

        for v in values:
            self.put(putter, v)
            gotten = getter()
            assert eq(gotten, v), (gotten, v)

        self.data.rewind()

        for v in values:
            vtype = self.data.next()
            assert vtype == ntype, vtype
            gotten = getter()
            assert eq(gotten, v), (gotten, v)

        encoded = self.data.encode()
        copy = Data(0)
        while encoded:
            n = copy.decode(encoded)
            encoded = encoded[n:]
        copy.rewind()

        cgetter = getattr(copy, "get_%s" % dtype)

        for v in values:
            vtype = copy.next()
            assert vtype == ntype, vtype
            gotten = cgetter()
            assert eq(gotten, v), (gotten, v)

    def _test_int(self, itype):
        self._test(itype, *self.int_values(itype))

    def testByte(self): self._test_int("byte")

    def testUbyte(self):
        self._test_int("ubyte")
        self.assertRaises(AssertionError, ubyte, -1)

    def testShort(self): self._test_int("short")

    def testUshort(self):
        self._test("ushort")
        self.assertRaises(AssertionError, ushort, -1)

    def testInt(self): self._test_int("int")

    def testUint(self):
        self._test_int("uint")
        self.assertRaises(AssertionError, uint, -1)

    def testLong(self): self._test_int("long")

    def testUlong(self):
        self._test_int("ulong")
        self.assertRaises(AssertionError, ulong, -1)

    def testString(self):
        self._test("string", "one", "two", "three", "this is a test", "")

    def testFloat(self):
        # we have to use a special comparison here because python
        # internally only uses doubles and converting between floats and
        # doubles is imprecise
        self._test("float", 0, 1, 2, 3, 0.1, 0.2, 0.3, -1, -2, -3, -0.1, -0.2, -0.3,
                   eq=lambda x, y: x - y < 0.000001)

    def testDouble(self):
        self._test("double", 0, 1, 2, 3, 0.1, 0.2, 0.3, -1, -2, -3, -0.1, -0.2, -0.3)

    def testBinary(self):
        self._test("binary", b"this", b"is", b"a", b"test", b"of" b"b\x00inary")

    def testSymbol(self):
        self._test("symbol", symbol("this is a symbol test"), symbol("bleh"), symbol("blah"))

    def testTimestamp(self):
        self._test("timestamp", timestamp(0), timestamp(12345), timestamp(1000000))

    def testChar(self):
        self._test("char", char('a'), char('b'), char('c'), char(u'\u20AC'))

    def testUUID(self):
        self._test("uuid", uuid4(), uuid4(), uuid4())

    def testDecimal32(self):
        self._test("decimal32", decimal32(0), decimal32(1), decimal32(2), decimal32(3), decimal32(4), decimal32(2**30))

    def testDecimal64(self):
        self._test("decimal64", decimal64(0), decimal64(1), decimal64(2), decimal64(3), decimal64(4), decimal64(2**60))

    def testDecimal128(self):
        self._test("decimal128", decimal128(b"fdsaasdf;lkjjkl;"), decimal128(b"x" * 16))

    def testCopy(self):
        self.data.put_described()
        self.data.enter()
        self.data.put_ulong(123)
        self.data.put_map()
        self.data.enter()
        self.data.put_string("pi")
        self.data.put_double(3.14159265359)

        dst = Data()
        dst.copy(self.data)

        copy = dst.format()
        orig = self.data.format()
        assert copy == orig, (copy, orig)

    def testCopyNested(self):
        nested = [1, 2, 3, [4, 5, 6], 7, 8, 9]
        self.data.put_object(nested)
        dst = Data()
        dst.copy(self.data)
        assert dst.format() == self.data.format()

    def testCopyNestedArray(self):
        nested = [Array(UNDESCRIBED, Data.LIST,
                        ["first", [Array(UNDESCRIBED, Data.INT, 1, 2, 3)]],
                        ["second", [Array(UNDESCRIBED, Data.INT, 1, 2, 3)]],
                        ["third", [Array(UNDESCRIBED, Data.INT, 1, 2, 3)]],
                        ),
                  "end"]
        self.data.put_object(nested)
        dst = Data()
        dst.copy(self.data)
        assert dst.format() == self.data.format()

    def testRoundTrip(self):
        obj = {symbol("key"): timestamp(1234),
               ulong(123): "blah",
               char("c"): "bleh",
               u"desc": Described(symbol("url"), u"http://example.org"),
               u"array": Array(UNDESCRIBED, Data.INT, 1, 2, 3),
               u"list": [1, 2, 3, None, 4],
               u"boolean": True}
        self.data.put_object(obj)
        enc = self.data.encode()
        data = Data()
        data.decode(enc)
        data.rewind()
        assert data.next()
        copy = data.get_object()
        assert copy == obj, (copy, obj)

    def testBuffer(self):
        try:
            self.data.put_object(buffer(b"foo"))
        except NameError:
            # python >= 3.0 does not have `buffer`
            return
        data = Data()
        data.decode(self.data.encode())
        data.rewind()
        assert data.next()
        assert data.type() == Data.BINARY
        assert data.get_object() == b"foo"

    def testMemoryView(self):
        self.data.put_object(memoryview(b"foo"))
        data = Data()
        data.decode(self.data.encode())
        data.rewind()
        assert data.next()
        assert data.type() == Data.BINARY
        assert data.get_object() == b"foo"

    def testLookup(self):
        obj = {symbol("key"): u"value",
               symbol("pi"): 3.14159,
               symbol("list"): [1, 2, 3, 4]}
        self.data.put_object(obj)
        self.data.rewind()
        self.data.next()
        self.data.enter()
        self.data.narrow()
        assert self.data.lookup("pi")
        assert self.data.get_object() == 3.14159
        self.data.rewind()
        assert self.data.lookup("key")
        assert self.data.get_object() == u"value"
        self.data.rewind()
        assert self.data.lookup("list")
        assert self.data.get_object() == [1, 2, 3, 4]
        self.data.widen()
        self.data.rewind()
        assert not self.data.lookup("pi")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

from proton import Url

from . import common


class UrlTest(common.Test):
    def assertEqual(self, a, b):
        assert a == b, "%s != %s" % (a, b)

    def assertNotEqual(self, a, b):
        assert a != b, "%s == %s" % (a, b)

    def assertUrl(self, u, scheme, username, password, host, port, path):
        self.assertEqual((u.scheme, u.username, u.password, u.host, u.port, u.path),
                         (scheme, username, password, host, port, path))

    def testUrl(self):
        url = Url('amqp://me:secret@myhost:1234/foobar')
        self.assertEqual(str(url), "amqp://me:secret@myhost:1234/foobar")
        self.assertUrl(url, 'amqp', 'me', 'secret', 'myhost', 1234, 'foobar')
        self.assertEqual(str(url), "amqp://me:secret@myhost:1234/foobar")

    def testIpV6Url(self):
        url = Url('amqp://me:secret@[::1]:1234/foobar')
        self.assertEqual(str(url), "amqp://me:secret@[::1]:1234/foobar")
        self.assertUrl(url, 'amqp', 'me', 'secret', '::1', 1234, 'foobar')
        self.assertEqual(str(url), "amqp://me:secret@[::1]:1234/foobar")

    def testDefaults(self):
        # Check that we allow None for scheme, port
        url = Url(username='me', password='secret', host='myhost', path='foobar', defaults=False)
        self.assertEqual(str(url), "//me:secret@myhost/foobar")
        self.assertUrl(url, None, 'me', 'secret', 'myhost', None, 'foobar')

        self.assertEqual(str(Url("amqp://me:secret@myhost/foobar")),
                         "amqp://me:secret@myhost:amqp/foobar")

        # Empty string vs. None for path
        self.assertEqual(Url("myhost/").path, "")
        assert Url("myhost", defaults=False).path is None

        # Expanding abbreviated url strings.
        for s, u in [
            ("", "amqp://0.0.0.0:amqp"),
            ("foo", "amqp://foo:amqp"),
            (":1234", "amqp://0.0.0.0:1234"),
            ("/path", "amqp://0.0.0.0:amqp/path"),
            ("user@host/topic://test", "amqp://user@host:amqp/topic://test"),
            ("user@host:3456", "amqp://user@host:3456"),
            ("user:pass@host/topic://test", "amqp://user:pass@host:amqp/topic://test")
        ]:
            self.assertEqual(str(Url(s)), u)

    def assertPort(self, port, portint, portstr):
        self.assertEqual((port, port), (portint, portstr))
        self.assertEqual((int(port), str(port)), (portint, portstr))

    def testPort(self):
        self.assertPort(Url.Port('amqp'), 5672, 'amqp')
        self.assertPort(Url.Port(5672), 5672, '5672')
        self.assertPort(Url.Port(5671), 5671, '5671')
        self.assertEqual(Url.Port(5671) + 1, 5672)  # Treat as int
        self.assertEqual(str(Url.Port(5672)), '5672')

        self.assertPort(Url.Port(Url.Port('amqp')), 5672, 'amqp')
        self.assertPort(Url.Port(Url.Port(5672)), 5672, '5672')

        try:
            Url.Port('xxx')
            assert False, "Expected ValueError"
        except ValueError:
            pass

        self.assertEqual(str(Url("host:amqp", defaults=False)), "//host:amqp")
        self.assertEqual(Url("host:amqp", defaults=False).port, 5672)

    def testArgs(self):
        u = Url("amqp://u:p@host:amqp/path", scheme='foo', host='bar', port=1234, path='garden', defaults=False)
        self.assertUrl(u, 'foo', 'u', 'p', 'bar', 1234, 'garden')
        u = Url(defaults=False)
        self.assertUrl(u, None, None, None, None, None, None)

    def assertRaises(self, exception, function, *args, **kwargs):
        try:
            function(*args, **kwargs)
            assert False, "Expected exception %s" % exception.__name__
        except exception:
            pass

    def testMissing(self):
        self.assertUrl(Url(defaults=False), None, None, None, None, None, None)
        self.assertUrl(Url('amqp://', defaults=False), 'amqp', None, None, None, None, None)
        self.assertUrl(Url('username@', defaults=False), None, 'username', None, None, None, None)
        self.assertUrl(Url(':pass@', defaults=False), None, '', 'pass', None, None, None)
        self.assertUrl(Url('host', defaults=False), None, None, None, 'host', None, None)
        self.assertUrl(Url(':1234', defaults=False), None, None, None, None, 1234, None)
        self.assertUrl(Url('/path', defaults=False), None, None, None, None, None, 'path')

        for s, full in [
            ('amqp://', 'amqp://'),
            ('username@', '//username@'),
            (':pass@', '//:pass@'),
            (':1234', '//:1234'),
            ('/path', '/path')
        ]:
            self.assertEqual(str(Url(s, defaults=False)), full)

        for s, full in [
                ('amqp://', 'amqp://0.0.0.0:amqp'),
                ('username@', 'amqp://username@0.0.0.0:amqp'),
                (':pass@', 'amqp://:pass@0.0.0.0:amqp'),
                (':1234', 'amqp://0.0.0.0:1234'),
                ('/path', 'amqp://0.0.0.0:amqp/path'),
                ('foo/path', 'amqp://foo:amqp/path'),
                (':1234/path', 'amqp://0.0.0.0:1234/path')
        ]:
            self.assertEqual(str(Url(s)), full)

    def testAmqps(self):
        # Scheme defaults
        self.assertEqual(str(Url("me:secret@myhost/foobar")),
                         "amqp://me:secret@myhost:amqp/foobar")
        # Correct port for amqps vs. amqps
        self.assertEqual(str(Url("amqps://me:secret@myhost/foobar")),
                         "amqps://me:secret@myhost:amqps/foobar")

        self.assertPort(Url.Port('amqps'), 5671, 'amqps')
        self.assertEqual(str(Url("host:amqps", defaults=False)), "//host:amqps")
        self.assertEqual(Url("host:amqps", defaults=False).port, 5671)

    def testEqual(self):
        self.assertEqual(Url("foo/path"), 'amqp://foo:amqp/path')
        self.assertEqual('amqp://foo:amqp/path', Url("foo/path"))
        self.assertEqual(Url("foo/path"), Url("foo/path"))
        self.assertNotEqual(Url("foo/path"), 'xamqp://foo:amqp/path')
        self.assertNotEqual('xamqp://foo:amqp/path', Url("foo/path"))
        self.assertNotEqual(Url("foo/path"), Url("bar/path"))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

from threading import Thread, Event
from uuid import uuid4

from proton import Message, Url, Array, UNDESCRIBED, Data, symbol, ConnectionException
from proton.handlers import MessagingHandler
from proton.reactor import Container
from proton.utils import SyncRequestResponse, BlockingConnection

from .common import Test, free_tcp_port
from .common import ensureCanTestExtendedSASL

CONNECTION_PROPERTIES = {u'connection': u'properties'}
OFFERED_CAPABILITIES = Array(UNDESCRIBED, Data.SYMBOL, symbol("O_one"), symbol("O_two"), symbol("O_three"))
DESIRED_CAPABILITIES = Array(UNDESCRIBED, Data.SYMBOL, symbol("D_one"), symbol("D_two"), symbol("D_three"))
ANONYMOUS = 'ANONYMOUS'
EXTERNAL = 'EXTERNAL'


class EchoServer(MessagingHandler, Thread):
    """
    Simple echo server that echos messages to their reply-to. Runs in a thread.
    Will only accept a single connection and shut down when that connection closes.
    """

    def __init__(self, url, timeout):
        MessagingHandler.__init__(self)
        Thread.__init__(self)
        self.daemon = True
        self.timeout = timeout
        self.url = url
        self.senders = {}
        self.container = None
        self.event = Event()

    def on_start(self, event):
        self.acceptor = event.container.listen(self.url)
        self.container = event.container
        self.event.set()

    def on_link_opening(self, event):
        if event.link.is_sender:
            if event.link.remote_source and event.link.remote_source.dynamic:
                event.link.source.address = str(uuid4())
                self.senders[event.link.source.address] = event.link

    def on_message(self, event):
        m = event.message
        sender = self.senders.get(m.reply_to)
        if sender:
            reply = Message(address=m.reply_to, body=m.body, correlation_id=m.correlation_id)
            sender.send(reply)

    def on_connection_closing(self, event):
        self.acceptor.close()

    def on_transport_error(self, event):
        self.acceptor.close()

    def run(self):
        Container(self).run()

    def wait(self):
        self.event.wait(self.timeout)


class ConnPropertiesServer(EchoServer):
    def __init__(self, url, timeout):
        EchoServer.__init__(self, url, timeout)
        self.properties_received = False
        self.offered_capabilities_received = False
        self.desired_capabilities_received = False

    def on_connection_opening(self, event):
        conn = event.connection

        if conn.remote_properties == CONNECTION_PROPERTIES:
            self.properties_received = True
        if conn.remote_offered_capabilities == OFFERED_CAPABILITIES:
            self.offered_capabilities_received = True
        if conn.remote_desired_capabilities == DESIRED_CAPABILITIES:
            self.desired_capabilities_received = True


class SyncRequestResponseTest(Test):
    """Test SyncRequestResponse"""

    def test_request_response(self):
        ensureCanTestExtendedSASL()

        def test(name, address="x"):
            for i in range(5):
                body = "%s%s" % (name, i)
                response = client.call(Message(address=address, body=body))
                self.assertEquals(response.address, client.reply_to)
                self.assertEquals(response.body, body)

        server = EchoServer(Url(host="127.0.0.1", port=free_tcp_port()), self.timeout)
        server.start()
        server.wait()
        connection = BlockingConnection(server.url, timeout=self.timeout)
        client = SyncRequestResponse(connection)
        try:
            test("foo")         # Simple request/resposne
        finally:
            client.connection.close()
        server.join(timeout=self.timeout)

    def test_connection_properties(self):
        ensureCanTestExtendedSASL()
        server = ConnPropertiesServer(Url(host="127.0.0.1", port=free_tcp_port()), timeout=self.timeout)
        server.start()
        server.wait()
        connection = BlockingConnection(server.url, timeout=self.timeout, properties=CONNECTION_PROPERTIES,
                                        offered_capabilities=OFFERED_CAPABILITIES, desired_capabilities=DESIRED_CAPABILITIES)
        client = SyncRequestResponse(connection)
        client.connection.close()
        server.join(timeout=self.timeout)
        self.assertEquals(server.properties_received, True)
        self.assertEquals(server.offered_capabilities_received, True)
        self.assertEquals(server.desired_capabilities_received, True)

    def test_allowed_mechs_external(self):
        # All this test does it make sure that if we pass allowed_mechs to BlockingConnection, it is actually used.
        port = free_tcp_port()
        server = ConnPropertiesServer(Url(host="127.0.0.1", port=port), timeout=self.timeout)
        server.start()
        server.wait()
        try:
            # This will cause an exception because we are specifying allowed_mechs as
            # EXTERNAL. The SASL handshake will fail because the server is not setup
            # to handle EXTERNAL
            connection = BlockingConnection(server.url, timeout=self.timeout, properties=CONNECTION_PROPERTIES,
                                            offered_capabilities=OFFERED_CAPABILITIES, desired_capabilities=DESIRED_CAPABILITIES, allowed_mechs=EXTERNAL)
            self.fail("Expected ConnectionException")
        except ConnectionException as e:
            self.assertTrue('amqp:unauthorized-access' in str(e), "expected unauthorized-access")
        server.join(timeout=self.timeout)

    def test_allowed_mechs_anonymous(self):
        # All this test does it make sure that if we pass allowed_mechs to BlockingConnection, it is actually used.
        server = ConnPropertiesServer(Url(host="127.0.0.1", port=free_tcp_port()), timeout=self.timeout)
        server.start()
        server.wait()
        # An ANONYMOUS allowed_mechs will work, anonymous connections are allowed by ConnPropertiesServer
        connection = BlockingConnection(server.url, timeout=self.timeout, properties=CONNECTION_PROPERTIES,
                                        offered_capabilities=OFFERED_CAPABILITIES, desired_capabilities=DESIRED_CAPABILITIES, allowed_mechs=ANONYMOUS)
        client = SyncRequestResponse(connection)
        client.connection.close()
        server.join(timeout=self.timeout)
        self.assertEquals(server.properties_received, True)
        self.assertEquals(server.offered_capabilities_received, True)
        self.assertEquals(server.desired_capabilities_received, True)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""
This module provides document parsing and transformation utilities for XML.
"""

from __future__ import absolute_import

import os
import sys
import xml.sax
import types
from xml.sax.handler import ErrorHandler
from xml.sax.xmlreader import InputSource

try:
    from io import StringIO
except ImportError:
    from cStringIO import StringIO

if sys.version_info[0] == 2:
    import types
    CLASS_TYPES = (type, types.ClassType)
else:
    CLASS_TYPES = (type,)

from . import dom
from . import transforms
from . import parsers


def transform(node, *args):
    result = node
    for t in args:
        if isinstance(t, CLASS_TYPES):
            t = t()
        result = result.dispatch(t)
    return result


class Resolver:

    def __init__(self, path):
        self.path = path

    def resolveEntity(self, publicId, systemId):
        for p in self.path:
            fname = os.path.join(p, systemId)
            if os.path.exists(fname):
                source = InputSource(systemId)
                source.setByteStream(open(fname))
                return source
        return InputSource(systemId)


def xml_parse(filename, path=()):
    h = parsers.XMLParser()
    p = xml.sax.make_parser()
    p.setContentHandler(h)
    p.setErrorHandler(ErrorHandler())
    p.setEntityResolver(Resolver(path))
    p.parse(filename)
    return h.parser.tree


def sexp(node):
    s = transforms.Sexp()
    node.dispatch(s)
    return s.out



#!/usr/bin/env python3
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import optparse
from proton.handlers import MessagingHandler
from proton.reactor import ApplicationEvent, Container, EventInjector
from db_common import Db


class Recv(MessagingHandler):
    def __init__(self, url, count):
        super(Recv, self).__init__(auto_accept=False)
        self.url = url
        self.delay = 0
        self.last_id = None
        self.expected = count
        self.received = 0
        self.accepted = 0
        self.db = Db("dst_db", EventInjector())

    def on_start(self, event):
        event.container.selectable(self.db.injector)
        e = ApplicationEvent("id_loaded")
        e.container = event.container
        self.db.get_id(e)

    def on_id_loaded(self, event):
        self.last_id = event.id
        event.container.create_receiver(self.url)

    def on_record_inserted(self, event):
        self.accept(event.delivery)
        self.accepted += 1
        if self.accepted == self.expected:
            event.connection.close()
            self.db.close()

    def on_message(self, event):
        id = int(event.message.id)
        if (not self.last_id) or id > self.last_id:
            if self.expected == 0 or self.received < self.expected:
                self.received += 1
                self.last_id = id
                self.db.insert(id, event.message.body, ApplicationEvent("record_inserted", delivery=event.delivery))
                print("inserted message %s" % id)
            else:
                self.release(event.delivery)
        else:
            self.accept(event.delivery)


parser = optparse.OptionParser(usage="usage: %prog [options]")
parser.add_option("-a", "--address", default="localhost:5672/examples",
                  help="address from which messages are received (default %default)")
parser.add_option("-m", "--messages", type="int", default=0,
                  help="number of messages to receive; 0 receives indefinitely (default %default)")
opts, args = parser.parse_args()

try:
    Container(Recv(opts.address, opts.messages)).run()
except KeyboardInterrupt:
    pass



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

import os
import gc
import traceback

from proton import *
from proton.reactor import Container

from . import common

CUSTOM = EventType("custom")


class HandlerTest(common.Test):
    def test_reactorHandlerCycling(self, n=0):

        class CustomHandler(Handler):
            UNSET = 999999999

            def __init__(self):
                self.offset = len(traceback.extract_stack())

            def on_reactor_init(self, event):
                self.depth = len(traceback.extract_stack())

            def reset(self):
                self.depth = self.UNSET

            @property
            def init_depth(self):
                d = self.depth - self.offset
                return d
        custom = CustomHandler()

        container = Container()
        container.handler = custom
        for i in range(n):
            h = container.handler
            container.handler = h
        custom.reset()
        container.run()
        assert custom.init_depth < 50, "Unexpectedly long traceback for a simple handler"

    def test_reactorHandlerCycling10k(self):
        self.test_reactorHandlerCycling(10000)

    def test_reactorHandlerCycling100(self):
        self.test_reactorHandlerCycling(100)

    def do_customEvent(self, reactor_handler, event_root):

        class CustomHandler:
            did_custom = False
            did_init = False

            def __init__(self, *handlers):
                self.handlers = handlers

            def on_reactor_init(self, event):
                self.did_init = True

            def on_custom(self, event):
                self.did_custom = True

        class CustomInvoker(CustomHandler):
            def on_reactor_init(self, event):
                h = event_root(event)
                event.dispatch(h, CUSTOM)
                self.did_init = True

        child = CustomInvoker()
        root = CustomHandler(child)

        container = Container()

        reactor_handler(container, root)
        container.run()
        assert root.did_init
        assert child.did_init
        assert root.did_custom
        assert child.did_custom

    def set_root(self, reactor, root):
        reactor.handler = root

    def add_root(self, reactor, root):
        reactor.handler.add(root)

    def append_root(self, reactor, root):
        reactor.handler.handlers.append(root)

    def event_root(self, event):
        return event.handler

    def event_reactor_handler(self, event):
        return event.reactor.handler

    def test_set_handler(self):
        self.do_customEvent(self.set_root, self.event_reactor_handler)

    def test_add_handler(self):
        self.do_customEvent(self.add_root, self.event_reactor_handler)

    def test_append_handler(self):
        self.do_customEvent(self.append_root, self.event_reactor_handler)

    def test_set_root(self):
        self.do_customEvent(self.set_root, self.event_root)

    def test_add_root(self):
        self.do_customEvent(self.add_root, self.event_root)

    def test_append_root(self):
        self.do_customEvent(self.append_root, self.event_root)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import optparse
import socket
import sys
import threading

import cproton

import proton.handlers
import proton.reactor
import proton.utils


class Broker(proton.handlers.MessagingHandler):
    def __init__(self, acceptor_url):
        # type: (str) -> None
        super(Broker, self).__init__()
        self.acceptor_url = acceptor_url

        self.acceptor = None
        self._acceptor_opened_event = threading.Event()

    def get_acceptor_sockname(self):
        # type: () -> (str, int)
        self._acceptor_opened_event.wait()
        if hasattr(self.acceptor, '_selectable'):  # proton 0.30.0+
            sockname = self.acceptor._selectable._delegate.getsockname()
        else:  # works in proton 0.27.0
            selectable = cproton.pn_cast_pn_selectable(self.acceptor._impl)
            fd = cproton.pn_selectable_get_fd(selectable)
            s = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
            sockname = s.getsockname()
        return sockname[:2]

    def on_start(self, event):
        self.acceptor = event.container.listen(self.acceptor_url)
        self._acceptor_opened_event.set()

    def on_link_opening(self, event):
        if event.link.is_sender:
            assert not event.link.remote_source.dynamic, "This cannot happen"
            event.link.source.address = event.link.remote_source.address
        elif event.link.remote_target.address:
            event.link.target.address = event.link.remote_target.address


def main():
    parser = optparse.OptionParser()
    parser.add_option("-b", dest="hostport", default="localhost:0", type="string",
                      help="port number to use")
    options, args = parser.parse_args()

    broker = Broker(options.hostport)
    container = proton.reactor.Container(broker)
    threading.Thread(target=container.run).start()
    print("{0}:{1}".format(*broker.get_acceptor_sockname()))
    sys.stdout.flush()


if __name__ == '__main__':
    main()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from typing import Optional, Union, Any


def isinteger(value: Any) -> bool:
    return isinstance(value, int)


def isstring(value: Any) -> bool:
    return isinstance(value, str)


class Constant(object):

    def __init__(self, name: str) -> None:
        self.name = name

    def __repr__(self) -> str:
        return self.name


def secs2millis(secs: Union[float, int]) -> int:
    return int(secs * 1000)


def millis2secs(millis: int) -> float:
    return float(millis) / 1000.0


def unicode2utf8(string: Optional[str]) -> Optional[str]:
    """Some Proton APIs expect a null terminated string. Convert python text
    types to UTF8 to avoid zero bytes introduced by other multi-byte encodings.
    This method will throw if the string cannot be converted.
    """
    if string is None:
        return None
    elif isinstance(string, str):
        # The swig binding converts py3 str -> utf8 char* and back automatically
        return string
    # Anything else illegal - specifically python3 bytes
    raise TypeError("Unrecognized string type: %r (%s)" % (string, type(string)))


def utf82unicode(string: Optional[Union[str, bytes]]) -> Optional[str]:
    """Convert C strings returned from proton-c into python unicode"""
    if string is None:
        return None
    elif isinstance(string, str):
        return string
    elif isinstance(string, bytes):
        return string.decode('utf8')
    raise TypeError("Unrecognized string type")



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import os
import subprocess
import time
import unittest


class Popen(subprocess.Popen):

    # We always use these options
    def __init__(self, args, **kwargs):
        # Add the module directory to the path to be able to find the examples
        path = f"{os.path.dirname(__file__)}{os.pathsep}{os.environ['PATH']}"
        super(Popen, self).\
            __init__(args,
                     shell=False,
                     stderr=subprocess.STDOUT,
                     stdout=subprocess.PIPE,
                     universal_newlines=True,
                     env={'PATH': path},
                     **kwargs)


# Like subprocess.run with our defaults but returning the Popen object
def run(args, **kwargs):
    p = Popen(args, **kwargs)
    p.wait()
    return p


def check_call(args, **kwargs):
    # Add the module directory to the path to be able to find the examples
    path = f"{os.path.dirname(__file__)}{os.pathsep}{os.environ['PATH']}"
    return subprocess.check_call(args, env={'PATH': path}, **kwargs)


class ExamplesTest(unittest.TestCase):
    def test_helloworld(self, example="helloworld.py"):
        with run([example]) as p:
            output = [l.strip() for l in p.stdout]
            self.assertEqual(output, ['Hello World!'])

    def test_helloworld_direct(self):
        self.test_helloworld('helloworld_direct.py')

    def test_helloworld_blocking(self):
        self.test_helloworld('helloworld_blocking.py')

    def test_helloworld_tornado(self):
        self.test_helloworld('helloworld_tornado.py')

    def test_helloworld_direct_tornado(self):
        self.test_helloworld('helloworld_direct_tornado.py')

    def test_simple_send_recv(self, recv='simple_recv.py', send='simple_send.py'):
        with Popen([recv]) as r:
            with run([send]):
                pass
            actual = [l.strip() for l in r.stdout]
            expected = ["{'sequence': %i}" % (i + 1,) for i in range(100)]
            self.assertEqual(actual, expected)

    def test_client_server(self, client=['client.py'], server=['server.py'], sleep=0):
        with Popen(server) as s:
            if sleep:
                time.sleep(sleep)
            with run(client) as c:
                actual = [l.strip() for l in c.stdout]
                inputs = ["Twas brillig, and the slithy toves",
                          "Did gire and gymble in the wabe.",
                          "All mimsy were the borogroves,",
                          "And the mome raths outgrabe."]
                expected = ["%s => %s" % (l, l.upper()) for l in inputs]
                self.assertEqual(actual, expected)
            s.terminate()

    def test_sync_client_server(self):
        self.test_client_server(client=['sync_client.py'])

    def test_client_server_tx(self):
        self.test_client_server(server=['server_tx.py'])

    def test_sync_client_server_tx(self):
        self.test_client_server(client=['sync_client.py'], server=['server_tx.py'])

    def test_client_server_direct(self):
        self.test_client_server(client=['client.py', '-a', 'localhost:8888/examples'],
                                server=['server_direct.py'], sleep=0.5)

    def test_sync_client_server_direct(self):
        self.test_client_server(client=['sync_client.py', '-a', 'localhost:8888/examples'],
                                server=['server_direct.py'], sleep=0.5)

    def test_db_send_recv(self):
        self.maxDiff = None
        # setup databases
        check_call(['db_ctrl.py', 'init', './src_db'])
        check_call(['db_ctrl.py', 'init', './dst_db'])
        with Popen(['db_ctrl.py', 'insert', './src_db'], stdin=subprocess.PIPE) as fill:
            for i in range(100):
                fill.stdin.write("Message-%i\n" % (i + 1))
            fill.stdin.close()

        # run send and recv
        with Popen(['db_recv.py', '-m', '100']) as r:
            with run(['db_send.py', '-m', '100']):
                pass
            r.wait()
            # verify output of receive
            actual = [l.strip() for l in r.stdout]
            expected = ["inserted message %i" % (i + 1) for i in range(100)]
            self.assertEqual(actual, expected)

        # verify state of databases
        with run(['db_ctrl.py', 'list', './dst_db']) as v:
            expected = ["(%i, 'Message-%i')" % (i + 1, i + 1) for i in range(100)]
            actual = [l.strip() for l in v.stdout]
            self.assertEqual(actual, expected)

    def test_tx_send_tx_recv(self):
        self.test_simple_send_recv(recv='tx_recv.py', send='tx_send.py')

    def test_simple_send_direct_recv(self):
        self.maxDiff = None
        with Popen(['direct_recv.py', '-a', 'localhost:8888']) as r:
            time.sleep(0.5)
            with run(['simple_send.py', '-a', 'localhost:8888']):
                pass
            r.wait()
            actual = [l.strip() for l in r.stdout]
            expected = ["{'sequence': %i}" % (i + 1,) for i in range(100)]
            self.assertEqual(actual, expected)

    def test_direct_send_simple_recv(self):
        with Popen(['direct_send.py', '-a', 'localhost:8888']):
            time.sleep(0.5)
            with run(['simple_recv.py', '-a', 'localhost:8888']) as r:
                actual = [l.strip() for l in r.stdout]
                expected = ["{'sequence': %i}" % (i + 1,) for i in range(100)]
                self.assertEqual(actual, expected)

    def test_selected_recv(self):
        with run(['colour_send.py']):
            pass

        with run(['selected_recv.py', '-m', '50']) as r:
            actual = [l.strip() for l in r.stdout]
            expected = ["green %i" % (i + 1) for i in range(100) if i % 2 == 0]
            self.assertEqual(actual, expected)

        with run(['simple_recv.py', '-m', '50']) as r:
            actual = [l.strip() for l in r.stdout]
            expected = ["red %i" % (i + 1) for i in range(100) if i % 2 == 1]
            self.assertEqual(actual, expected)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import absolute_import

import os

from proton import *

from . import common
from .common import Skipped, pump


def _testpath(file):
    """ Set the full path to the certificate,keyfile, etc. for the test.
    """
    if os.name == "nt":
        if file.find("private-key") != -1:
            # The private key is not in a separate store
            return None
        # Substitute pkcs#12 equivalent for the CA/key store
        if file.endswith(".pem"):
            file = file[:-4] + ".p12"
    return os.path.join(os.path.dirname(__file__),
                        "ssl_db/%s" % file)


class SslTest(common.Test):

    def __init__(self, *args):
        common.Test.__init__(self, *args)
        self._testpath = _testpath

    def setUp(self):
        if not common.isSSLPresent():
            raise Skipped("No SSL libraries found.")
        self.server_domain = SSLDomain(SSLDomain.MODE_SERVER)
        self.client_domain = SSLDomain(SSLDomain.MODE_CLIENT)

    def tearDown(self):
        self.server_domain = None
        self.client_domain = None

    class SslTestConnection(object):
        """ Represents a single SSL connection.
        """

        def __init__(self, domain=None, mode=Transport.CLIENT,
                     session_details=None, conn_hostname=None,
                     ssl_peername=None):
            if not common.isSSLPresent():
                raise Skipped("No SSL libraries found.")

            self.ssl = None
            self.domain = domain
            self.transport = Transport(mode)
            self.connection = Connection()
            if conn_hostname:
                self.connection.hostname = conn_hostname
            if domain:
                self.ssl = SSL(self.transport, self.domain, session_details)
                if ssl_peername:
                    self.ssl.peer_hostname = ssl_peername
            # bind last, after all configuration complete:
            self.transport.bind(self.connection)

    def _pump(self, ssl_client, ssl_server, buffer_size=1024):
        pump(ssl_client.transport, ssl_server.transport, buffer_size)

    def _do_handshake(self, client, server):
        """ Attempt to connect client to server. Will throw a TransportException if the SSL
        handshake fails.
        """
        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        if client.transport.closed:
            return
        assert client.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_anonymous_cipher(self):
        if os.name == "nt":
            raise Skipped("Windows SChannel lacks anonymous cipher support.")
        """ With no configuration at all, the client default
        VERIFY_PEER_NAME should preclude anonymous cipher TLS negotiation.
        """
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        # check that no SSL connection exists
        assert not server.ssl.cipher_name()
        assert not client.ssl.protocol_name()

        # client.transport.trace(Transport.TRACE_DRV)
        # server.transport.trace(Transport.TRACE_DRV)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)

        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT

    def test_simple_anonymous(self):
        if os.name == "nt":
            raise Skipped("Windows SChannel lacks anonymous cipher support.")
        """ The simplest SSL configuration using anonymous
        ciphers.
        """
        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        # check that no SSL connection exists
        assert not server.ssl.cipher_name()
        assert not client.ssl.protocol_name()

        # client.transport.trace(Transport.TRACE_DRV)
        # server.transport.trace(Transport.TRACE_DRV)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)

        # now SSL should be active
        assert server.ssl.cipher_name() is not None
        assert client.ssl.protocol_name() is not None

        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_ssl_with_small_buffer(self):
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()

        small_buffer_size = 1
        self._pump(client, server, small_buffer_size)

        assert client.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_server_certificate_fail(self):
        """ Test that default configured clients cannot connect to a server that has
        a certificate configured.
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)

        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT

    def test_server_certificate_no_verify(self):
        """ Test that anonymous clients can still connect to a server that has
        a certificate configured.
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_server_authentication(self):
        """ Simple SSL connection with authentication of the server
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")

        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_certificate_fingerprint_and_subfields(self):
        if os.name == "nt":
            raise Skipped("Windows support for certificate fingerprint and subfield not implemented yet")

        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   self._testpath("ca-certificate.pem"))
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)

        # give the client a certificate, but let's not require server authentication
        self.client_domain.set_credentials(self._testpath("client-certificate1.pem"),
                                           self._testpath("client-private-key1.pem"),
                                           "client-password")
        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)

        # Test the subject subfields
        self.assertEqual("Client", server.ssl.get_cert_organization())
        self.assertEqual("Dev", server.ssl.get_cert_organization_unit())
        self.assertEqual("ST", server.ssl.get_cert_state_or_province())
        self.assertEqual("US", server.ssl.get_cert_country())
        self.assertEqual("City", server.ssl.get_cert_locality_or_city())
        self.assertEqual("O=Server,CN=A1.Good.Server.domain.com", client.ssl.get_cert_subject())
        self.assertEqual("O=Client,CN=127.0.0.1,C=US,ST=ST,L=City,OU=Dev", server.ssl.get_cert_subject())

        self.assertEqual("1762688fb3c1c7908690efa158e89924a8e739b6", server.ssl.get_cert_fingerprint_sha1())
        self.assertEqual("f5890052d5ad3f38f68eae9027b30fc4c3f09758d59d8f19e3f192e29e41708d",
                         server.ssl.get_cert_fingerprint_sha256())
        self.assertEqual("5d2b1bff39df99d101040348be015970e3da1d0ad610902dc78bba24555aae16395fd1342e26b89b422c304c32b48913000b485933720773d033aaffc561b3e9",
                         server.ssl.get_cert_fingerprint_sha512())
        self.assertEqual("246ea13c8a549e8c0b0cc6490d81b34b", server.ssl.get_cert_fingerprint_md5())

        # Test the various fingerprint algorithms
        self.assertEqual("d28c0ae17c370c269bd680ea3bcc523ea5da544e", client.ssl.get_cert_fingerprint_sha1())
        self.assertEqual("c460a601f77b77ec59480955574a227c309805dd36dbf866f42d2dce3fd4757c",
                         client.ssl.get_cert_fingerprint_sha256())
        self.assertEqual("37c28f451105b1979e2ea62d4a38c86e158ad345894b5016662bdd1913f48764bd71deb4b4de4ce22559828634357dcaea1832dd58327dfe5b0bc368ecbeee4c",
                         client.ssl.get_cert_fingerprint_sha512())
        self.assertEqual("16b075688b82c40ce5b03c984f20286b", client.ssl.get_cert_fingerprint_md5())

        self.assertEqual(None, client.ssl.get_cert_fingerprint(21, SSL.SHA1))  # Should be at least 41
        self.assertEqual(None, client.ssl.get_cert_fingerprint(50, SSL.SHA256))  # Should be at least 65
        self.assertEqual(None, client.ssl.get_cert_fingerprint(128, SSL.SHA512))  # Should be at least 129
        self.assertEqual(None, client.ssl.get_cert_fingerprint(10, SSL.MD5))  # Should be at least 33
        self.assertEqual(None, client.ssl._get_cert_subject_unknown_subfield())

        self.assertNotEqual(None, client.ssl.get_cert_fingerprint(50, SSL.SHA1))  # Should be at least 41
        self.assertNotEqual(None, client.ssl.get_cert_fingerprint(70, SSL.SHA256))  # Should be at least 65
        self.assertNotEqual(None, client.ssl.get_cert_fingerprint(130, SSL.SHA512))  # Should be at least 129
        self.assertNotEqual(None, client.ssl.get_cert_fingerprint(35, SSL.MD5))  # Should be at least 33
        self.assertEqual(None, client.ssl._get_cert_fingerprint_unknown_hash_alg())

    def test_client_authentication(self):
        """ Force the client to authenticate.
        """
        # note: when requesting client auth, the server _must_ send its
        # certificate, so make sure we configure one!
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   self._testpath("ca-certificate.pem"))
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)

        # give the client a certificate, but let's not require server authentication
        self.client_domain.set_credentials(self._testpath("client-certificate.pem"),
                                           self._testpath("client-private-key.pem"),
                                           "client-password")
        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)

        assert client.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_client_authentication_fail_bad_cert(self):
        """ Ensure that the server can detect a bad client certificate.
        """
        # note: when requesting client auth, the server _must_ send its
        # certificate, so make sure we configure one!
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   self._testpath("ca-certificate.pem"))
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)

        self.client_domain.set_credentials(self._testpath("bad-server-certificate.pem"),
                                           self._testpath("bad-server-private-key.pem"),
                                           "server-password")
        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT

    def test_client_authentication_fail_no_cert(self):
        """ Ensure that the server will fail a client that does not provide a
        certificate.
        """
        # note: when requesting client auth, the server _must_ send its
        # certificate, so make sure we configure one!
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   self._testpath("ca-certificate.pem"))
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)

        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT

    def test_client_server_authentication(self):
        """ Require both client and server to mutually identify themselves.
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   self._testpath("ca-certificate.pem"))

        self.client_domain.set_credentials(self._testpath("client-certificate.pem"),
                                           self._testpath("client-private-key.pem"),
                                           "client-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_server_only_authentication(self):
        """ Client verifies server, but server does not verify client.
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)

        self.client_domain.set_credentials(self._testpath("client-certificate.pem"),
                                           self._testpath("client-private-key.pem"),
                                           "client-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_bad_server_certificate(self):
        """ A server with a self-signed certificate that is not trusted by the
        client.  The client should reject the server.
        """
        self.server_domain.set_credentials(self._testpath("bad-server-certificate.pem"),
                                           self._testpath("bad-server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)

        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT

        del server
        del client

        # now re-try with a client that does not require peer verification
        self.client_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)

        client = SslTest.SslTestConnection(self.client_domain)
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_allow_unsecured_client_which_connects_unsecured(self):
        """ Server allows an unsecured client to connect if configured.
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   self._testpath("ca-certificate.pem"))
        # allow unsecured clients on this connection
        self.server_domain.allow_unsecured_client()
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)

        # non-ssl connection
        client = SslTest.SslTestConnection()

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert server.ssl.protocol_name() is None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_allow_unsecured_client_which_connects_secured(self):
        """ As per test_allow_unsecured_client_which_connects_unsecured
            but client actually uses SSL
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   self._testpath("ca-certificate.pem"))

        self.client_domain.set_credentials(self._testpath("client-certificate.pem"),
                                           self._testpath("client-private-key.pem"),
                                           "client-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        # allow unsecured clients on this connection
        # self.server_domain.allow_unsecured_client()

        # client uses ssl. Server should detect this.
        client = SslTest.SslTestConnection(self.client_domain)
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert server.ssl.protocol_name() is not None
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_disallow_unsecured_client(self):
        """ Non-SSL Client is disallowed from connecting to server.
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)

        # non-ssl connection
        client = SslTest.SslTestConnection()

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT

    def test_session_resume(self):
        """ Test resume of client session.
        """
        if os.name == "nt":
            raise Skipped("Windows SChannel session resume not yet implemented.")

        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_peer_authentication(SSLDomain.ANONYMOUS_PEER)

        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        # details will be used in initial and subsequent connections to allow session to be resumed
        initial_session_details = SSLSessionDetails("my-session-id")

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain, session_details=initial_session_details)

        # bring up the connection and store its state
        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert client.ssl.protocol_name() is not None

        # cleanly shutdown the connection
        client.connection.close()
        server.connection.close()
        self._pump(client, server)

        # destroy the existing clients
        del client
        del server

        # now create a new set of connections, use last session id
        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        # provide the details of the last session, allowing it to be resumed
        client = SslTest.SslTestConnection(self.client_domain, session_details=initial_session_details)

        # client.transport.trace(Transport.TRACE_DRV)
        # server.transport.trace(Transport.TRACE_DRV)

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert server.ssl.protocol_name() is not None
        if(API_LANGUAGE == "C"):
            assert client.ssl.resume_status() == SSL.RESUME_REUSED
        else:
            # Java gives no way to check whether a previous session has been resumed
            pass

        client.connection.close()
        server.connection.close()
        self._pump(client, server)

        # now try to resume using an unknown session-id, expect resume to fail
        # and a new session is negotiated

        del client
        del server

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(
            self.client_domain, session_details=SSLSessionDetails("some-other-session-id"))

        client.connection.open()
        server.connection.open()
        self._pump(client, server)
        assert server.ssl.protocol_name() is not None
        if(API_LANGUAGE == "C"):
            assert client.ssl.resume_status() == SSL.RESUME_NEW

        client.connection.close()
        server.connection.close()
        self._pump(client, server)

    def test_multiple_sessions(self):
        """ Test multiple simultaneous active SSL sessions with bi-directional
        certificate verification, shared across two domains.
        """
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.server_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.server_domain.set_peer_authentication(SSLDomain.VERIFY_PEER,
                                                   self._testpath("ca-certificate.pem"))

        self.client_domain.set_credentials(self._testpath("client-certificate.pem"),
                                           self._testpath("client-private-key.pem"),
                                           "client-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER)

        max_count = 100
        sessions = [(SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER),
                     SslTest.SslTestConnection(self.client_domain)) for x in
                    range(max_count)]
        for s in sessions:
            s[0].connection.open()
            self._pump(s[0], s[1])

        for s in sessions:
            s[1].connection.open()
            self._pump(s[1], s[0])
            assert s[0].ssl.cipher_name() is not None
            assert s[1].ssl.cipher_name() == s[0].ssl.cipher_name()

        for s in sessions:
            s[1].connection.close()
            self._pump(s[0], s[1])

        for s in sessions:
            s[0].connection.close()
            self._pump(s[1], s[0])

    def test_server_hostname_authentication(self):
        """ Test authentication of the names held in the server's certificate
        against various configured hostnames.
        """
        if os.name == "nt":
            raise Skipped("PROTON-1057: disable temporarily on Windows.")

        # Check the CommonName matches (case insensitive).
        # Assumes certificate contains "CN=A1.Good.Server.domain.com"
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.ssl.peer_hostname = "a1.good.server.domain.com"
        assert client.ssl.peer_hostname == "a1.good.server.domain.com"
        self._do_handshake(client, server)
        del server
        del client
        self.tearDown()

        # Should fail on CN name mismatch:
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.ssl.peer_hostname = "A1.Good.Server.domain.comX"
        self._do_handshake(client, server)
        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT
        del server
        del client
        self.tearDown()

        # Wildcarded Certificate
        # Assumes:
        #   1) certificate contains Server Alternate Names:
        #        "alternate.name.one.com" and "another.name.com"
        #   2) certificate has wildcarded CommonName "*.prefix*.domain.com"
        #

        # Pass: match an alternate
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
                                           self._testpath("server-wc-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.ssl.peer_hostname = "alternate.Name.one.com"
        self._do_handshake(client, server)
        del client
        del server
        self.tearDown()

        # Pass: match an alternate
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
                                           self._testpath("server-wc-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.ssl.peer_hostname = "ANOTHER.NAME.COM"
        self._do_handshake(client, server)
        del client
        del server
        self.tearDown()

        # Pass: match the pattern
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
                                           self._testpath("server-wc-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.ssl.peer_hostname = "SOME.PREfix.domain.COM"
        self._do_handshake(client, server)
        del client
        del server
        self.tearDown()

        # Pass: match the pattern
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
                                           self._testpath("server-wc-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.ssl.peer_hostname = "FOO.PREfixZZZ.domain.com"
        self._do_handshake(client, server)
        del client
        del server
        self.tearDown()

        # Fail: must match prefix on wildcard
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
                                           self._testpath("server-wc-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.ssl.peer_hostname = "FOO.PREfi.domain.com"
        self._do_handshake(client, server)
        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT
        del server
        del client
        self.tearDown()

        # Fail: leading wildcards are not optional
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
                                           self._testpath("server-wc-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        client.ssl.peer_hostname = "PREfix.domain.COM"
        self._do_handshake(client, server)
        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT
        self.tearDown()

        # Pass: ensure that the user can give an alternate name that overrides
        # the connection's configured hostname
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
                                           self._testpath("server-wc-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain,
                                           conn_hostname="This.Name.Does.not.Match",
                                           ssl_peername="alternate.name.one.com")
        self._do_handshake(client, server)
        del client
        del server
        self.tearDown()

        # Pass: ensure that the hostname supplied by the connection is used if
        # none has been specified for the SSL instance
        self.setUp()
        self.server_domain.set_credentials(self._testpath("server-certificate.pem"),
                                           self._testpath("server-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain,
                                           conn_hostname="a1.good.server.domain.com")
        self._do_handshake(client, server)
        del client
        del server
        self.tearDown()

    def test_server_hostname_authentication_2(self):
        """Initially separated from test_server_hostname_authentication
        above to force Windows checking and sidestep PROTON-1057 exclusion.
        """

        # Fail for a null peer name.
        self.server_domain.set_credentials(self._testpath("server-wc-certificate.pem"),
                                           self._testpath("server-wc-private-key.pem"),
                                           "server-password")
        self.client_domain.set_trusted_ca_db(self._testpath("ca-certificate.pem"))
        self.client_domain.set_peer_authentication(SSLDomain.VERIFY_PEER_NAME)

        server = SslTest.SslTestConnection(self.server_domain, mode=Transport.SERVER)
        client = SslTest.SslTestConnection(self.client_domain)

        # Next line results in an eventual pn_ssl_set_peer_hostname(client.ssl._ssl, None)
        client.ssl.peer_hostname = None
        self._do_handshake(client, server)
        assert client.transport.closed
        assert server.transport.closed
        assert client.connection.state & Endpoint.REMOTE_UNINIT
        assert server.connection.state & Endpoint.REMOTE_UNINIT
        self.tearDown()

    def test_defaults_messenger_app(self):
        """ Test an SSL connection using the Messenger apps (no certificates)
        """
        if os.name == "nt":
            raise Skipped("Windows SChannel lacks anonymous cipher support.")
        port = common.free_tcp_ports()[0]

        receiver = common.MessengerReceiverC()
        receiver.subscriptions = ["amqps://~0.0.0.0:%s" % port]
        receiver.receive_count = 1
        receiver.timeout = self.timeout
        receiver.start()

        sender = common.MessengerSenderC()
        sender.targets = ["amqps://0.0.0.0:%s/X" % port]
        sender.send_count = 1
        sender.timeout = self.timeout
        sender.start()
        sender.wait()
        assert sender.status() == 0, "Command '%s' failed" % str(sender.cmdline())

        receiver.wait()
        assert receiver.status() == 0, "Command '%s' failed" % str(receiver.cmdline())

    def test_server_authentication_messenger_app(self):
        """ Test an SSL authentication using the Messenger apps.
        """
        port = common.free_tcp_ports()[0]

        receiver = common.MessengerReceiverC()
        receiver.subscriptions = ["amqps://~0.0.0.0:%s" % port]
        receiver.receive_count = 1
        receiver.timeout = self.timeout
        # Note hack - by default we use the client-certificate for the
        # _server_ because the client-certificate's common name field
        # is "127.0.0.1", which will match the target address used by
        # the sender.
        receiver.certificate = self._testpath("client-certificate.pem")
        receiver.privatekey = self._testpath("client-private-key.pem")
        receiver.password = "client-password"
        receiver.start()

        sender = common.MessengerSenderC()
        sender.targets = ["amqps://127.0.0.1:%s/X" % port]
        sender.send_count = 1
        sender.timeout = self.timeout
        sender.ca_db = self._testpath("ca-certificate.pem")
        sender.start()
        sender.wait()
        assert sender.status() == 0, "Command '%s' failed" % str(sender.cmdline())

        receiver.wait()
        assert receiver.status() == 0, "Command '%s' failed" % str(receiver.cmdline())

    def test_singleton(self):
        """Verify that only a single instance of SSL can exist per Transport"""
        transport = Transport()
        ssl1 = SSL(transport, self.client_domain)
        ssl2 = transport.ssl(self.client_domain)
        ssl3 = transport.ssl(self.client_domain)
        assert ssl1 is ssl2
        assert ssl1 is ssl3
        transport = Transport()
        ssl1 = transport.ssl(self.client_domain)
        ssl2 = SSL(transport, self.client_domain)
        assert ssl1 is ssl2
        # catch attempt to re-configure existing SSL
        try:
            ssl3 = SSL(transport, self.server_domain)
            assert False, "Expected error did not occur!"
        except SSLException:
            pass



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License
#

"""
PROTON-2116 Memory leak in python client
PROTON-2192 Memory leak in Python client on Windows
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gc
import logging
import os
import platform
import subprocess
import sys
import threading
import time
import unittest
import uuid

import proton.handlers
import proton.reactor
import proton.utils


logger = logging.getLogger(__name__)


class ReconnectingTestClient:
    def __init__(self, hostport):
        # type: (str) -> None
        self.hostport = hostport

        self.object_counts = []
        self.done = threading.Event()

    def count_objects(self, message):
        # type: (str) -> None
        gc.collect()
        n = len(gc.get_objects())
        if message == "loop":
            self.object_counts.append(n)
        logger.debug("Message %s, Count %d", message, n)

    def run(self):
        ADDR = "testing123"
        HEARTBEAT = 5
        SLEEP = 5

        recv = None
        conn = None
        for _ in range(3):
            subscribed = False
            while not subscribed:
                try:
                    conn = proton.utils.BlockingConnection(self.hostport, ssl_domain=None, heartbeat=HEARTBEAT)
                    recv = conn.create_receiver(ADDR, name=str(uuid.uuid4()), dynamic=False, options=None)
                    subscribed = True
                except Exception as e:
                    logger.info("received exception %s on connect/subscribe, retry", e)
                    time.sleep(0.5)

            self.count_objects("loop")
            logger.debug("connected")
            while subscribed:
                try:
                    recv.receive(SLEEP)
                except proton.Timeout:
                    pass
                except Exception as e:
                    logger.info(e)
                    try:
                        recv.close()
                        recv = None
                    except:
                        self.count_objects("link close() failed")
                        pass
                    try:
                        conn.close()
                        conn = None
                        self.count_objects("conn closed")
                    except:
                        self.count_objects("conn close() failed")
                        pass
                    subscribed = False
        self.done.set()


class Proton2116Test(unittest.TestCase):
    @unittest.skipIf(platform.system() == 'Windows', "PROTON-2192: The issue is not resolved on Windows")
    def test_blocking_connection_object_leak(self):
        """Kills and restarts broker repeatedly, while client is reconnecting.

        The value of `gc.get_objects()` should not keep increasing in the client.

        These are the automated reproduction steps for PROTON-2116"""
        gc.collect()

        thread = None
        client = None

        host_port = ""  # random on first broker startup
        broker_process = None

        while not client or not client.done.is_set():
            try:
                params = []
                if host_port:
                    params = ['-b', host_port]
                cwd = os.path.dirname(__file__)
                broker_process = subprocess.Popen(
                    args=[sys.executable,
                          os.path.join(cwd, 'broker_PROTON_2116_blocking_connection_object_leak.py')] + params,
                    stdout=subprocess.PIPE,
                    universal_newlines=True,
                )
                host_port = broker_process.stdout.readline()

                if not client:
                    client = ReconnectingTestClient(host_port)
                    thread = threading.Thread(target=client.run)
                    thread.start()

                time.sleep(3)
            finally:
                if broker_process:
                    broker_process.kill()
                    broker_process.wait()
                    broker_process.stdout.close()
            time.sleep(0.3)

        thread.join()

        logger.info("client.object_counts:", client.object_counts)

        # drop first value, it is usually different (before counts settle)
        object_counts = client.object_counts[1:]

        diffs = [c - object_counts[0] for c in object_counts]
        self.assertEqual([0] * 2, diffs, "Object counts should not be increasing")


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import optparse
from proton import Url
from proton.reactor import Container
from proton.handlers import MessagingHandler, TransactionHandler


class TxRecv(MessagingHandler, TransactionHandler):
    def __init__(self, url, messages, batch_size):
        super(TxRecv, self).__init__(prefetch=0, auto_accept=False)
        self.url = Url(url)
        self.expected = messages
        self.batch_size = batch_size
        self.current_batch = 0
        self.committed = 0

    def on_start(self, event):
        self.container = event.container
        self.conn = self.container.connect(self.url)
        self.receiver = self.container.create_receiver(self.conn, self.url.path)
        self.container.declare_transaction(self.conn, handler=self)
        self.transaction = None

    def on_message(self, event):
        print(event.message.body)
        self.transaction.accept(event.delivery)
        self.current_batch += 1
        if self.current_batch == self.batch_size:
            self.transaction.commit()
            self.transaction = None

    def on_transaction_declared(self, event):
        self.receiver.flow(self.batch_size)
        self.transaction = event.transaction

    def on_transaction_committed(self, event):
        self.committed += self.current_batch
        self.current_batch = 0
        if self.expected == 0 or self.committed < self.expected:
            self.container.declare_transaction(self.conn, handler=self)
        else:
            event.connection.close()

    def on_disconnected(self, event):
        self.current_batch = 0


parser = optparse.OptionParser(usage="usage: %prog [options]")
parser.add_option("-a", "--address", default="localhost:5672/examples",
                  help="address from which messages are received (default %default)")
parser.add_option("-m", "--messages", type="int", default=100,
                  help="number of messages to receive; 0 receives indefinitely (default %default)")
parser.add_option("-b", "--batch-size", type="int", default=10,
                  help="number of messages in each transaction (default %default)")
opts, args = parser.parse_args()

try:
    Container(TxRecv(opts.address, opts.messages, opts.batch_size)).run()
except KeyboardInterrupt:
    pass



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""
Simple DOM for both SGML and XML documents.
"""

from __future__ import division
from __future__ import generators
from __future__ import nested_scopes
from __future__ import absolute_import

import sys
if sys.version_info[0] == 2:
    STRING_TYPES = (basestring,)
else:
    STRING_TYPES = (str,)


class Container:

    def __init__(self):
        self.children = []

    def add(self, child):
        child.parent = self
        self.children.append(child)

    def extend(self, children):
        for child in children:
            child.parent = self
            self.children.append(child)


class Component:

    def __init__(self):
        self.parent = None

    def index(self):
        if self.parent:
            return self.parent.children.index(self)
        else:
            return 0

    def _line(self, file, line, column):
        self.file = file
        self.line = line
        self.column = column


class DispatchError(Exception):

    def __init__(self, scope, f):
        msg = "no such attribute"


class Dispatcher:

    def is_type(self, type):
        cls = self
        while cls is not None:
            if cls.type == type:
                return True
            cls = cls.base
        return False

    def dispatch(self, f, attrs=""):
        cls = self
        while cls is not None:
            if hasattr(f, cls.type):
                return getattr(f, cls.type)(self)
            else:
                cls = cls.base

        cls = self
        while cls is not None:
            if attrs:
                sep = ", "
                if cls.base is None:
                    sep += "or "
            else:
                sep = ""
            attrs += "%s'%s'" % (sep, cls.type)
            cls = cls.base

        raise AttributeError("'%s' object has no attribute %s" %
                             (f.__class__.__name__, attrs))


class Node(Container, Component, Dispatcher):

    type = "node"
    base = None

    def __init__(self):
        Container.__init__(self)
        Component.__init__(self)
        self.query = Query([self])

    def __getitem__(self, name):
        for nd in self.query[name]:
            return nd

    def text(self):
        from . import transforms
        return self.dispatch(transforms.Text())

    def tag(self, name, *attrs, **kwargs):
        t = Tag(name, *attrs, **kwargs)
        self.add(t)
        return t

    def data(self, s):
        d = Data(s)
        self.add(d)
        return d

    def entity(self, s):
        e = Entity(s)
        self.add(e)
        return e


class Tree(Node):

    type = "tree"
    base = Node


class Tag(Node):

    type = "tag"
    base = Node

    def __init__(self, _name, *attrs, **kwargs):
        Node.__init__(self)
        self.name = _name
        self.attrs = list(attrs)
        self.attrs.extend(kwargs.items())
        self.singleton = False

    def get_attr(self, name):
        for k, v in self.attrs:
            if name == k:
                return v

    def _idx(self, attr):
        idx = 0
        for k, v in self.attrs:
            if k == attr:
                return idx
            idx += 1
        return None

    def set_attr(self, name, value):
        idx = self._idx(name)
        if idx is None:
            self.attrs.append((name, value))
        else:
            self.attrs[idx] = (name, value)

    def dispatch(self, f):
        try:
            attr = "do_" + self.name
            method = getattr(f, attr)
        except AttributeError:
            return Dispatcher.dispatch(self, f, "'%s'" % attr)
        return method(self)


class Leaf(Component, Dispatcher):

    type = "leaf"
    base = None

    def __init__(self, data):
        assert isinstance(data, STRING_TYPES)
        self.data = data


class Data(Leaf):
    type = "data"
    base = Leaf


class Entity(Leaf):
    type = "entity"
    base = Leaf


class Character(Leaf):
    type = "character"
    base = Leaf


class Comment(Leaf):
    type = "comment"
    base = Leaf

###################
## Query Classes ##
###########################################################################


class Adder:

    def __add__(self, other):
        return Sum(self, other)


class Sum(Adder):

    def __init__(self, left, right):
        self.left = left
        self.right = right

    def __iter__(self):
        for x in self.left:
            yield x
        for x in self.right:
            yield x


class View(Adder):

    def __init__(self, source):
        self.source = source


class Filter(View):

    def __init__(self, predicate, source):
        View.__init__(self, source)
        self.predicate = predicate

    def __iter__(self):
        for nd in self.source:
            if self.predicate(nd):
                yield nd


class Flatten(View):

    def __iter__(self):
        sources = [iter(self.source)]
        while sources:
            try:
                nd = next(sources[-1])
                if isinstance(nd, Tree):
                    sources.append(iter(nd.children))
                else:
                    yield nd
            except StopIteration:
                sources.pop()


class Children(View):

    def __iter__(self):
        for nd in self.source:
            for child in nd.children:
                yield child


class Attributes(View):

    def __iter__(self):
        for nd in self.source:
            for a in nd.attrs:
                yield a


class Values(View):

    def __iter__(self):
        for name, value in self.source:
            yield value


def flatten_path(path):
    if isinstance(path, STRING_TYPES):
        for part in path.split("/"):
            yield part
    elif callable(path):
        yield path
    else:
        for p in path:
            for fp in flatten_path(p):
                yield fp


class Query(View):

    def __iter__(self):
        for nd in self.source:
            yield nd

    def __getitem__(self, path):
        query = self.source
        for p in flatten_path(path):
            if callable(p):
                select = Query
                pred = p
                source = query
            elif isinstance(p, STRING_TYPES):
                if p[0] == "@":
                    select = Values
                    pred = lambda x, n=p[1:]: x[0] == n
                    source = Attributes(query)
                elif p[0] == "#":
                    select = Query
                    pred = lambda x, t=p[1:]: x.is_type(t)
                    source = Children(query)
                else:
                    select = Query
                    def pred(x, n=p): return isinstance(x, Tag) and x.name == n
                    source = Flatten(Children(query))
            else:
                raise ValueError(p)
            query = select(Filter(pred, source))

        return query



#!/usr/bin/env python3
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import optparse
from proton import Message
from proton.handlers import MessagingHandler
from proton.reactor import Container


class Send(MessagingHandler):
    def __init__(self, url, messages):
        super(Send, self).__init__()
        self.url = url
        self.sent = 0
        self.confirmed = 0
        self.total = messages

    def on_start(self, event):
        event.container.create_sender(self.url)

    def on_sendable(self, event):
        while event.sender.credit and self.sent < self.total:
            msg = Message(id=(self.sent + 1), body={'sequence': (self.sent + 1)})
            event.sender.send(msg)
            self.sent += 1

    def on_accepted(self, event):
        self.confirmed += 1
        if self.confirmed == self.total:
            print("all messages confirmed")
            event.connection.close()

    def on_disconnected(self, event):
        self.sent = self.confirmed


parser = optparse.OptionParser(usage="usage: %prog [options]",
                               description="Send messages to the supplied address.")
parser.add_option("-a", "--address", default="localhost:5672/examples",
                  help="address to which messages are sent (default %default)")
parser.add_option("-m", "--messages", type="int", default=100,
                  help="number of messages to send (default %default)")
opts, args = parser.parse_args()

try:
    Container(Send(opts.address, opts.messages)).run()
except KeyboardInterrupt:
    pass
